"""Control a `ProtocolEngine` without async/await."""

from typing import cast, List, Optional, Dict

from opentrons_shared_data.pipette.dev_types import PipetteNameType
from opentrons_shared_data.labware.dev_types import LabwareUri
from opentrons_shared_data.labware.labware_definition import LabwareDefinition

from opentrons.commands.protocol_commands import comment as make_legacy_comment_command
from opentrons.types import MountType
from opentrons.hardware_control.modules.types import ThermocyclerStep

from .. import commands
from ..state import StateView
from ..types import (
    DeckPoint,
    DeckSlotLocation,
    LabwareLocation,
    LabwareMovementStrategy,
    ModuleModel,
    WellLocation,
    DropTipWellLocation,
    LabwareOffsetVector,
    MotorAxis,
    Liquid,
    NozzleLayoutConfigurationType,
    AddressableOffsetVector,
)
from .transports import ChildThreadTransport


class SyncClient:
    """Control a `ProtocolEngine` without async/await.

    Normally, `ProtocolEngine` provides an async/await interface, like this:

    ```
    aspirate_result = await protocol_engine.add_and_execute_command(aspirate_command)
    dispense_result = await protocol_engine.add_and_execute_command(dispense_command)
    ```

    But we sometimes want to control it with plain old non-async blocking method calls.
    To accomplish that, this class adapts `ProtocolEngine`'s interface into this:

    ```
    aspirate_result = sync_client.aspirate(...)
    dispense_result = sync_client.dispense(...)
    ```

    This is intended to help implement the Python Protocol API, which is all non-async.
    """

    def __init__(self, transport: ChildThreadTransport) -> None:
        """Initialize the `SyncClient`.

        Params:
            transport: The interface for the new `SyncClient` to use to
                communicate with the `ProtocolEngine`.
        """
        self._transport = transport

    @property
    def state(self) -> StateView:
        """Get a view of the engine's state."""
        return self._transport.state

    def add_labware_definition(self, definition: LabwareDefinition) -> LabwareUri:
        """Add a labware definition to the engine."""
        return self._transport.call_method(
            "add_labware_definition",
            definition=definition,
        )

    def add_addressable_area(self, addressable_area_name: str) -> None:
        """Add an addressable area to the engine's state."""
        self._transport.call_method(
            "add_addressable_area", addressable_area_name=addressable_area_name
        )

    def add_liquid(
        self, name: str, color: Optional[str], description: Optional[str]
    ) -> Liquid:
        """Add a liquid to the engine."""
        return self._transport.call_method("add_liquid", name=name, color=color, description=description)  # type: ignore[no-any-return]

    def reset_tips(self, labware_id: str) -> None:
        """Reset a labware's tip tracking state.."""
        self._transport.call_method(
            "reset_tips",
            labware_id=labware_id,
        )

    def set_pipette_movement_speed(
        self, pipette_id: str, speed: Optional[float]
    ) -> None:
        """Set the speed of a pipette's X/Y/Z movements. Does not affect plunger speed.

        None will use the hardware API's default.
        """
        self._transport.call_method(
            "set_pipette_movement_speed",
            pipette_id=pipette_id,
            speed=speed,
        )

    def load_labware(
        self,
        location: LabwareLocation,
        load_name: str,
        namespace: str,
        version: int,
        display_name: Optional[str] = None,
    ) -> commands.LoadLabwareResult:
        """Execute a LoadLabware command and return the result."""
        request = commands.LoadLabwareCreate(
            params=commands.LoadLabwareParams(
                location=location,
                loadName=load_name,
                namespace=namespace,
                version=version,
                displayName=display_name,
            )
        )
        result = self._transport.execute_command(request=request)

        return cast(commands.LoadLabwareResult, result)

    # TODO (spp, 2022-12-14): https://opentrons.atlassian.net/browse/RLAB-237
    def move_labware(
        self,
        labware_id: str,
        new_location: LabwareLocation,
        strategy: LabwareMovementStrategy,
        pick_up_offset: Optional[LabwareOffsetVector],
        drop_offset: Optional[LabwareOffsetVector],
    ) -> commands.MoveLabwareResult:
        """Execute a MoveLabware command and return the result."""
        request = commands.MoveLabwareCreate(
            params=commands.MoveLabwareParams(
                labwareId=labware_id,
                newLocation=new_location,
                strategy=strategy,
                pickUpOffset=pick_up_offset,
                dropOffset=drop_offset,
            )
        )
        result = self._transport.execute_command(request=request)

        return cast(commands.MoveLabwareResult, result)

    def load_pipette(
        self,
        pipette_name: PipetteNameType,
        mount: MountType,
    ) -> commands.LoadPipetteResult:
        """Execute a LoadPipette command and return the result."""
        request = commands.LoadPipetteCreate(
            params=commands.LoadPipetteParams(mount=mount, pipetteName=pipette_name)
        )
        result = self._transport.execute_command(request=request)

        return cast(commands.LoadPipetteResult, result)

    def move_to_well(
        self,
        pipette_id: str,
        labware_id: str,
        well_name: str,
        well_location: WellLocation,
        minimum_z_height: Optional[float],
        force_direct: bool,
        speed: Optional[float],
    ) -> commands.MoveToWellResult:
        """Execute a MoveToWell command and return the result."""
        request = commands.MoveToWellCreate(
            params=commands.MoveToWellParams(
                pipetteId=pipette_id,
                labwareId=labware_id,
                wellName=well_name,
                wellLocation=well_location,
                forceDirect=force_direct,
                minimumZHeight=minimum_z_height,
                speed=speed,
            )
        )
        result = self._transport.execute_command(request=request)

        return cast(commands.MoveToWellResult, result)

    def move_to_addressable_area(
        self,
        pipette_id: str,
        addressable_area_name: str,
        offset: AddressableOffsetVector,
        minimum_z_height: Optional[float],
        force_direct: bool,
        speed: Optional[float],
    ) -> commands.MoveToAddressableAreaResult:
        """Execute a MoveToAddressableArea command and return the result."""
        request = commands.MoveToAddressableAreaCreate(
            params=commands.MoveToAddressableAreaParams(
                pipetteId=pipette_id,
                addressableAreaName=addressable_area_name,
                offset=offset,
                forceDirect=force_direct,
                minimumZHeight=minimum_z_height,
                speed=speed,
            )
        )
        result = self._transport.execute_command(request=request)

        return cast(commands.MoveToAddressableAreaResult, result)

    def move_to_addressable_area_for_drop_tip(
        self,
        pipette_id: str,
        addressable_area_name: str,
        offset: AddressableOffsetVector,
        minimum_z_height: Optional[float],
        force_direct: bool,
        speed: Optional[float],
        alternate_drop_location: Optional[bool],
        ignore_tip_configuration: Optional[bool] = True,
    ) -> commands.MoveToAddressableAreaForDropTipResult:
        """Execute a MoveToAddressableArea command and return the result."""
        request = commands.MoveToAddressableAreaForDropTipCreate(
            params=commands.MoveToAddressableAreaForDropTipParams(
                pipetteId=pipette_id,
                addressableAreaName=addressable_area_name,
                offset=offset,
                forceDirect=force_direct,
                minimumZHeight=minimum_z_height,
                speed=speed,
                alternateDropLocation=alternate_drop_location,
                ignoreTipConfiguration=ignore_tip_configuration,
            )
        )
        result = self._transport.execute_command(request=request)

        return cast(commands.MoveToAddressableAreaForDropTipResult, result)

    def move_to_coordinates(
        self,
        pipette_id: str,
        coordinates: DeckPoint,
        minimum_z_height: Optional[float],
        force_direct: bool,
        speed: Optional[float],
    ) -> commands.MoveToCoordinatesResult:
        """Execute a MoveToCoordinates command and return the result."""
        request = commands.MoveToCoordinatesCreate(
            params=commands.MoveToCoordinatesParams(
                pipetteId=pipette_id,
                coordinates=coordinates,
                minimumZHeight=minimum_z_height,
                forceDirect=force_direct,
                speed=speed,
            )
        )
        result = self._transport.execute_command(request=request)

        return cast(commands.MoveToCoordinatesResult, result)

    def load_module(
        self,
        model: ModuleModel,
        location: DeckSlotLocation,
    ) -> commands.LoadModuleResult:
        """Execute a LoadModule command and return the result."""
        request = commands.LoadModuleCreate(
            params=commands.LoadModuleParams(model=model, location=location)
        )
        result = self._transport.execute_command(request=request)

        return cast(commands.LoadModuleResult, result)

    def pick_up_tip(
        self,
        pipette_id: str,
        labware_id: str,
        well_name: str,
        well_location: WellLocation,
    ) -> commands.PickUpTipResult:
        """Execute a PickUpTip command and return the result."""
        request = commands.PickUpTipCreate(
            params=commands.PickUpTipParams(
                pipetteId=pipette_id,
                labwareId=labware_id,
                wellName=well_name,
                wellLocation=well_location,
            )
        )
        result = self._transport.execute_command(request=request)

        return cast(commands.PickUpTipResult, result)

    def drop_tip(
        self,
        pipette_id: str,
        labware_id: str,
        well_name: str,
        well_location: DropTipWellLocation,
        home_after: Optional[bool],
        alternateDropLocation: Optional[bool],
    ) -> commands.DropTipResult:
        """Execute a DropTip command and return the result."""
        request = commands.DropTipCreate(
            params=commands.DropTipParams(
                pipetteId=pipette_id,
                labwareId=labware_id,
                wellName=well_name,
                wellLocation=well_location,
                homeAfter=home_after,
                alternateDropLocation=alternateDropLocation,
            )
        )
        result = self._transport.execute_command(request=request)
        return cast(commands.DropTipResult, result)

    def drop_tip_in_place(
        self,
        pipette_id: str,
        home_after: Optional[bool],
    ) -> commands.DropTipInPlaceResult:
        """Execute a DropTipInPlace command and return the result."""
        request = commands.DropTipInPlaceCreate(
            params=commands.DropTipInPlaceParams(
                pipetteId=pipette_id,
                homeAfter=home_after,
            )
        )
        result = self._transport.execute_command(request=request)
        return cast(commands.DropTipInPlaceResult, result)

    def configure_for_volume(
        self, pipette_id: str, volume: float
    ) -> commands.ConfigureForVolumeResult:
        """Execute a ConfigureForVolume command."""
        request = commands.ConfigureForVolumeCreate(
            params=commands.ConfigureForVolumeParams(
                pipetteId=pipette_id, volume=volume
            )
        )
        result = self._transport.execute_command(request=request)
        return cast(commands.ConfigureForVolumeResult, result)

    def prepare_to_aspirate(self, pipette_id: str) -> commands.PrepareToAspirateResult:
        """Execute a PrepareToAspirate command."""
        request = commands.PrepareToAspirateCreate(
            params=commands.PrepareToAspirateParams(pipetteId=pipette_id)
        )
        result = self._transport.execute_command(request=request)
        return cast(commands.PrepareToAspirateResult, result)

    def configure_nozzle_layout(
        self,
        pipette_id: str,
        configuration_params: NozzleLayoutConfigurationType,
    ) -> commands.ConfigureNozzleLayoutResult:
        """Execute a ConfigureForVolume command."""
        request = commands.ConfigureNozzleLayoutCreate(
            params=commands.ConfigureNozzleLayoutParams(
                pipetteId=pipette_id, configurationParams=configuration_params
            )
        )
        result = self._transport.execute_command(request=request)
        return cast(commands.ConfigureNozzleLayoutResult, result)

    def aspirate(
        self,
        pipette_id: str,
        labware_id: str,
        well_name: str,
        well_location: WellLocation,
        volume: float,
        flow_rate: float,
    ) -> commands.AspirateResult:
        """Execute an ``Aspirate`` command and return the result."""
        request = commands.AspirateCreate(
            params=commands.AspirateParams(
                pipetteId=pipette_id,
                labwareId=labware_id,
                wellName=well_name,
                wellLocation=well_location,
                volume=volume,
                flowRate=flow_rate,
            )
        )
        result = self._transport.execute_command(request=request)

        return cast(commands.AspirateResult, result)

    def aspirate_in_place(
        self,
        pipette_id: str,
        volume: float,
        flow_rate: float,
    ) -> commands.AspirateInPlaceResult:
        """Execute an ``AspirateInPlace`` command and return the result."""
        request = commands.AspirateInPlaceCreate(
            params=commands.AspirateInPlaceParams(
                pipetteId=pipette_id,
                volume=volume,
                flowRate=flow_rate,
            )
        )
        result = self._transport.execute_command(request=request)

        return cast(commands.AspirateInPlaceResult, result)

    def dispense(
        self,
        pipette_id: str,
        labware_id: str,
        well_name: str,
        well_location: WellLocation,
        volume: float,
        flow_rate: float,
        push_out: Optional[float],
    ) -> commands.DispenseResult:
        """Execute a ``Dispense`` command and return the result."""
        request = commands.DispenseCreate(
            params=commands.DispenseParams(
                pipetteId=pipette_id,
                labwareId=labware_id,
                wellName=well_name,
                wellLocation=well_location,
                volume=volume,
                flowRate=flow_rate,
                pushOut=push_out,
            )
        )
        result = self._transport.execute_command(request=request)
        return cast(commands.DispenseResult, result)

    def dispense_in_place(
        self,
        pipette_id: str,
        volume: float,
        flow_rate: float,
        push_out: Optional[float],
    ) -> commands.DispenseInPlaceResult:
        """Execute a ``DispenseInPlace`` command and return the result."""
        request = commands.DispenseInPlaceCreate(
            params=commands.DispenseInPlaceParams(
                pipetteId=pipette_id,
                volume=volume,
                flowRate=flow_rate,
                pushOut=push_out,
            )
        )
        result = self._transport.execute_command(request=request)
        return cast(commands.DispenseInPlaceResult, result)

    def blow_out(
        self,
        pipette_id: str,
        labware_id: str,
        well_name: str,
        well_location: WellLocation,
        flow_rate: float,
    ) -> commands.BlowOutResult:
        """Execute a ``BlowOut`` command and return the result."""
        request = commands.BlowOutCreate(
            params=commands.BlowOutParams(
                pipetteId=pipette_id,
                labwareId=labware_id,
                wellName=well_name,
                wellLocation=well_location,
                flowRate=flow_rate,
            )
        )
        result = self._transport.execute_command(request=request)
        return cast(commands.BlowOutResult, result)

    def blow_out_in_place(
        self,
        pipette_id: str,
        flow_rate: float,
    ) -> commands.BlowOutInPlaceResult:
        """Execute a ``BlowOutInPlace`` command and return the result."""
        request = commands.BlowOutInPlaceCreate(
            params=commands.BlowOutInPlaceParams(
                pipetteId=pipette_id,
                flowRate=flow_rate,
            )
        )
        result = self._transport.execute_command(request=request)
        return cast(commands.BlowOutInPlaceResult, result)

    def touch_tip(
        self,
        pipette_id: str,
        labware_id: str,
        well_name: str,
        well_location: WellLocation,
        radius: float,
        speed: float,
    ) -> commands.TouchTipResult:
        """Execute a ``Touch Tip`` command and return the result."""
        request = commands.TouchTipCreate(
            params=commands.TouchTipParams(
                pipetteId=pipette_id,
                labwareId=labware_id,
                wellName=well_name,
                wellLocation=well_location,
                radius=radius,
                speed=speed,
            )
        )
        result = self._transport.execute_command(request=request)
        return cast(commands.TouchTipResult, result)

    def wait_for_duration(
        self, seconds: float, message: Optional[str]
    ) -> commands.WaitForDurationResult:
        """Execute a ``waitForDuration`` command and return the result."""
        request = commands.WaitForDurationCreate(
            params=commands.WaitForDurationParams(seconds=seconds, message=message)
        )
        result = self._transport.execute_command(request=request)
        return cast(commands.WaitForDurationResult, result)

    def wait_for_resume(self, message: Optional[str]) -> commands.WaitForResumeResult:
        """Execute a `WaitForResume` command and return the result."""
        request = commands.WaitForResumeCreate(
            params=commands.WaitForResumeParams(message=message)
        )
        result = self._transport.execute_command(request=request)
        return cast(commands.WaitForResumeResult, result)

    def comment(self, message: str) -> commands.CustomResult:
        """Execute a comment command and return the result."""
        # TODO(mm, 2022-11-09): Protocol Engine doesn't yet have a proper comment
        # command. So, we use a legacy-style command wrapped inside a Protocol Engine
        # CustomCommand. The Opentrons App knows how to render this in its run log
        # because this is what we used to do for PAPIv2 commands in general.
        #
        # When Protocol Engine has a proper comment command, we should use it here.
        legacy_comment_command = make_legacy_comment_command(msg=message)

        class LegacyCommentCustomParams(commands.CustomParams):
            legacyCommandType: str
            legacyCommandText: str

        request = commands.CustomCreate(
            params=LegacyCommentCustomParams(
                # This matches how LegacyCommandWrapper handles comments coming from
                # protocols running under the older non-ProtocolEngine core.
                legacyCommandType=legacy_comment_command["name"],
                legacyCommandText=legacy_comment_command["payload"]["text"],
            )
        )
        result = self._transport.execute_command(request=request)
        return cast(commands.CustomResult, result)

    def set_rail_lights(self, on: bool) -> commands.SetRailLightsResult:
        """Execute a ``setRailLights`` command and return the result."""
        request = commands.SetRailLightsCreate(
            params=commands.SetRailLightsParams(on=on)
        )
        result = self._transport.execute_command(request=request)
        return cast(commands.SetRailLightsResult, result)

    def magnetic_module_engage(
        self, module_id: str, engage_height: float
    ) -> commands.magnetic_module.EngageResult:
        """Execute a ``MagneticModuleEngage`` command and return the result."""
        request = commands.magnetic_module.EngageCreate(
            params=commands.magnetic_module.EngageParams(
                moduleId=module_id, height=engage_height
            )
        )
        result = self._transport.execute_command(request=request)
        return cast(commands.magnetic_module.EngageResult, result)

    def magnetic_module_disengage(
        self, module_id: str
    ) -> commands.magnetic_module.DisengageResult:
        """Execute a ``MagneticModuleDisengage`` command and return the result."""
        request = commands.magnetic_module.DisengageCreate(
            params=commands.magnetic_module.DisengageParams(moduleId=module_id)
        )
        result = self._transport.execute_command(request=request)
        return cast(commands.magnetic_module.DisengageResult, result)

    def thermocycler_set_target_lid_temperature(
        self, module_id: str, celsius: float
    ) -> commands.thermocycler.SetTargetLidTemperatureResult:
        """Execute a `thermocycler/setTargetLidTemperature` command and return the result."""
        request = commands.thermocycler.SetTargetLidTemperatureCreate(
            params=commands.thermocycler.SetTargetLidTemperatureParams(
                moduleId=module_id, celsius=celsius
            )
        )
        result = self._transport.execute_command(request=request)
        return cast(commands.thermocycler.SetTargetLidTemperatureResult, result)

    def thermocycler_set_target_block_temperature(
        self,
        module_id: str,
        celsius: float,
        block_max_volume: Optional[float],
        hold_time_seconds: Optional[float],
    ) -> commands.thermocycler.SetTargetBlockTemperatureResult:
        """Execute a `thermocycler/setTargetLidTemperature` command and return the result."""
        request = commands.thermocycler.SetTargetBlockTemperatureCreate(
            params=commands.thermocycler.SetTargetBlockTemperatureParams(
                moduleId=module_id,
                celsius=celsius,
                blockMaxVolumeUl=block_max_volume,
                holdTimeSeconds=hold_time_seconds,
            )
        )
        result = self._transport.execute_command(request=request)
        return cast(commands.thermocycler.SetTargetBlockTemperatureResult, result)

    def thermocycler_wait_for_lid_temperature(
        self, module_id: str
    ) -> commands.thermocycler.WaitForLidTemperatureResult:
        """Execute a `thermocycler/waitForLidTemperature` command and return the result."""
        request = commands.thermocycler.WaitForLidTemperatureCreate(
            params=commands.thermocycler.WaitForLidTemperatureParams(moduleId=module_id)
        )
        result = self._transport.execute_command(request=request)
        return cast(commands.thermocycler.WaitForLidTemperatureResult, result)

    def thermocycler_wait_for_block_temperature(
        self, module_id: str
    ) -> commands.thermocycler.WaitForBlockTemperatureResult:
        """Execute a `thermocycler/waitForBlockTemperature` command and return the result."""
        request = commands.thermocycler.WaitForBlockTemperatureCreate(
            params=commands.thermocycler.WaitForBlockTemperatureParams(
                moduleId=module_id
            )
        )
        result = self._transport.execute_command(request=request)
        return cast(commands.thermocycler.WaitForBlockTemperatureResult, result)

    def thermocycler_run_profile(
        self,
        module_id: str,
        steps: List[ThermocyclerStep],
        block_max_volume: Optional[float],
    ) -> commands.thermocycler.RunProfileResult:
        """Execute a `thermocycler/runProfile` command and return the result."""
        request = commands.thermocycler.RunProfileCreate(
            params=commands.thermocycler.RunProfileParams(
                moduleId=module_id,
                profile=[
                    commands.thermocycler.RunProfileStepParams(
                        celsius=step["temperature"],
                        holdSeconds=step["hold_time_seconds"],
                    )
                    for step in steps
                ],
                blockMaxVolumeUl=block_max_volume,
            )
        )
        result = self._transport.execute_command(request=request)
        return cast(commands.thermocycler.RunProfileResult, result)

    def thermocycler_deactivate_block(
        self, module_id: str
    ) -> commands.thermocycler.DeactivateBlockResult:
        """Execute a `thermocycler/deactivateBlock` command and return the result."""
        request = commands.thermocycler.DeactivateBlockCreate(
            params=commands.thermocycler.DeactivateBlockParams(moduleId=module_id)
        )
        result = self._transport.execute_command(request=request)
        return cast(commands.thermocycler.DeactivateBlockResult, result)

    def thermocycler_deactivate_lid(
        self, module_id: str
    ) -> commands.thermocycler.DeactivateLidResult:
        """Execute a `thermocycler/deactivateLid` command and return the result."""
        request = commands.thermocycler.DeactivateLidCreate(
            params=commands.thermocycler.DeactivateLidParams(moduleId=module_id)
        )
        result = self._transport.execute_command(request=request)
        return cast(commands.thermocycler.DeactivateLidResult, result)

    def thermocycler_open_lid(
        self, module_id: str
    ) -> commands.thermocycler.OpenLidResult:
        """Execute a `thermocycler/openLid` command and return the result."""
        request = commands.thermocycler.OpenLidCreate(
            params=commands.thermocycler.OpenLidParams(moduleId=module_id)
        )
        result = self._transport.execute_command(request=request)
        return cast(commands.thermocycler.OpenLidResult, result)

    def thermocycler_close_lid(
        self, module_id: str
    ) -> commands.thermocycler.CloseLidResult:
        """Execute a `thermocycler/closeLid` command and return the result."""
        request = commands.thermocycler.CloseLidCreate(
            params=commands.thermocycler.CloseLidParams(moduleId=module_id)
        )
        result = self._transport.execute_command(request=request)
        return cast(commands.thermocycler.CloseLidResult, result)

    def heater_shaker_set_target_temperature(
        self, module_id: str, celsius: float
    ) -> commands.heater_shaker.SetTargetTemperatureResult:
        """Execute a `heaterShaker/setTargetTemperature` command and return the result."""
        request = commands.heater_shaker.SetTargetTemperatureCreate(
            params=commands.heater_shaker.SetTargetTemperatureParams(
                moduleId=module_id, celsius=celsius
            )
        )
        result = self._transport.execute_command(request=request)
        return cast(commands.heater_shaker.SetTargetTemperatureResult, result)

    def heater_shaker_wait_for_temperature(
        self,
        module_id: str,
    ) -> commands.heater_shaker.WaitForTemperatureResult:
        """Execute a `heaterShaker/waitForTemperature` command and return the result."""
        request = commands.heater_shaker.WaitForTemperatureCreate(
            params=commands.heater_shaker.WaitForTemperatureParams(
                moduleId=module_id,
            )
        )
        result = self._transport.execute_command(request=request)
        return cast(commands.heater_shaker.WaitForTemperatureResult, result)

    def heater_shaker_set_and_wait_for_shake_speed(
        self, module_id: str, rpm: float
    ) -> commands.heater_shaker.SetAndWaitForShakeSpeedResult:
        """Execute a `heaterShaker/setAndWaitForShakeSpeed` command and return the result."""
        request = commands.heater_shaker.SetAndWaitForShakeSpeedCreate(
            params=commands.heater_shaker.SetAndWaitForShakeSpeedParams(
                moduleId=module_id, rpm=rpm
            )
        )
        result = self._transport.execute_command(request=request)
        return cast(commands.heater_shaker.SetAndWaitForShakeSpeedResult, result)

    def heater_shaker_open_labware_latch(
        self, module_id: str
    ) -> commands.heater_shaker.OpenLabwareLatchResult:
        """Execute a `heaterShaker/openLabwareLatch` command and return the result."""
        request = commands.heater_shaker.OpenLabwareLatchCreate(
            params=commands.heater_shaker.OpenLabwareLatchParams(moduleId=module_id)
        )
        result = self._transport.execute_command(request=request)
        return cast(commands.heater_shaker.OpenLabwareLatchResult, result)

    def heater_shaker_close_labware_latch(
        self, module_id: str
    ) -> commands.heater_shaker.CloseLabwareLatchResult:
        """Execute a `heaterShaker/closeLabwareLatch` command and return the result."""
        request = commands.heater_shaker.CloseLabwareLatchCreate(
            params=commands.heater_shaker.CloseLabwareLatchParams(moduleId=module_id)
        )
        result = self._transport.execute_command(request=request)
        return cast(commands.heater_shaker.CloseLabwareLatchResult, result)

    def heater_shaker_deactivate_shaker(
        self, module_id: str
    ) -> commands.heater_shaker.DeactivateShakerResult:
        """Execute a `heaterShaker/deactivateShaker` command and return the result."""
        request = commands.heater_shaker.DeactivateShakerCreate(
            params=commands.heater_shaker.DeactivateShakerParams(moduleId=module_id)
        )
        result = self._transport.execute_command(request=request)
        return cast(commands.heater_shaker.DeactivateShakerResult, result)

    def heater_shaker_deactivate_heater(
        self, module_id: str
    ) -> commands.heater_shaker.DeactivateHeaterResult:
        """Execute a `heaterShaker/deactivateHeater` command and return the result."""
        request = commands.heater_shaker.DeactivateHeaterCreate(
            params=commands.heater_shaker.DeactivateHeaterParams(moduleId=module_id)
        )
        result = self._transport.execute_command(request=request)
        return cast(commands.heater_shaker.DeactivateHeaterResult, result)

    def temperature_module_set_target_temperature(
        self, module_id: str, celsius: float
    ) -> commands.temperature_module.SetTargetTemperatureResult:
        """Execute a `temperatureModule/setTargetTemperature` command and return the result."""
        request = commands.temperature_module.SetTargetTemperatureCreate(
            params=commands.temperature_module.SetTargetTemperatureParams(
                moduleId=module_id, celsius=celsius
            ),
        )
        result = self._transport.execute_command(request=request)
        return cast(commands.temperature_module.SetTargetTemperatureResult, result)

    def temperature_module_wait_for_target_temperature(
        self, module_id: str, celsius: Optional[float]
    ) -> commands.temperature_module.WaitForTemperatureResult:
        """Execute a `temperatureModule/waitForTemperature` command and return the result."""
        request = commands.temperature_module.WaitForTemperatureCreate(
            params=commands.temperature_module.WaitForTemperatureParams(
                moduleId=module_id, celsius=celsius
            ),
        )
        result = self._transport.execute_command(request=request)
        return cast(commands.temperature_module.WaitForTemperatureResult, result)

    def temperature_module_deactivate(
        self, module_id: str
    ) -> commands.temperature_module.DeactivateTemperatureResult:
        """Execute a `temperatureModule/deactivate` command and return the result."""
        request = commands.temperature_module.DeactivateTemperatureCreate(
            params=commands.temperature_module.DeactivateTemperatureParams(
                moduleId=module_id
            ),
        )
        result = self._transport.execute_command(request=request)
        return cast(commands.temperature_module.DeactivateTemperatureResult, result)

    def home(self, axes: Optional[List[MotorAxis]]) -> commands.HomeResult:
        """Execute a `home` command and return the result."""
        request = commands.HomeCreate(params=commands.HomeParams(axes=axes))
        result = self._transport.execute_command(request=request)
        return cast(commands.HomeResult, result)

    def load_liquid(
        self, labware_id: str, liquid_id: str, volume_by_well: Dict[str, float]
    ) -> commands.LoadLiquidResult:
        """Execute a load_liquid command and return the result."""
        request = commands.LoadLiquidCreate(
            params=commands.LoadLiquidParams(
                labwareId=labware_id, liquidId=liquid_id, volumeByWell=volume_by_well
            )
        )
        result = self._transport.execute_command(request=request)
        return cast(commands.LoadLiquidResult, result)



"""Safely pass values between threads and async tasks."""


from __future__ import annotations

from collections import deque
from threading import Condition
from typing import AsyncIterable, Deque, Generic, Iterable, TypeVar

from anyio.to_thread import run_sync


_T = TypeVar("_T")


class ThreadAsyncQueue(Generic[_T]):
    """A queue to safely pass values of type `_T` between threads and async tasks.

    All methods are safe to call concurrently from any thread or task.

    Compared to queue.Queue:

    * This class lets you close the queue to signal that no more values will be added,
      which makes common producer/consumer patterns easier.
      (This is like Golang channels and AnyIO memory object streams.)
    * This class has built-in support for async consumers.

    Compared to asyncio.Queue and AnyIO memory object streams:

    * You can use this class to communicate between async tasks and threads
      without the threads having to wait for the event loop to be free
      every time they access the queue.
    """

    def __init__(self) -> None:
        """Initialize the queue."""
        self._is_closed = False
        self._deque: Deque[_T] = deque()
        self._condition = Condition()

    def put(self, value: _T) -> None:
        """Add a value to the back of the queue.

        Returns immediately, without blocking. The queue can grow without bound.

        Raises:
            QueueClosed: If the queue is already closed.
        """
        with self._condition:
            if self._is_closed:
                raise QueueClosed("Can't add more values when queue is already closed.")
            else:
                self._deque.append(value)
                self._condition.notify()

    def get(self) -> _T:
        """Remove and return the value at the front of the queue.

        If the queue is empty, this blocks until a new value is available.
        If you're calling from an async task, use one of the async methods instead
        to avoid blocking the event loop.

        Raises:
            QueueClosed: If all values have been consumed
                and the queue has been closed with `done_putting()`.
        """
        with self._condition:
            while True:
                if len(self._deque) > 0:
                    return self._deque.popleft()
                elif self._is_closed:
                    raise QueueClosed("Queue closed; no more items to get.")
                else:
                    # We don't have anything to return.
                    # Wait for something to change, then check again.
                    self._condition.wait()

    def get_until_closed(self) -> Iterable[_T]:
        """Remove and return values from the front of the queue until it's closed.

        Example:
            for value in queue.get_until_closed():
                print(value)
        """
        while True:
            try:
                yield self.get()
            except QueueClosed:
                break

    async def get_async(self) -> _T:
        """Like `get()`, except yield to the event loop while waiting.

        Warning:
            A waiting `get_async()` won't be interrupted by an async cancellation.
            The proper way to interrupt a waiting `get_async()`
            is to close the queue, just like you have to do with `get()`.
        """
        return await run_sync(
            self.get,
            # We keep `cancellable` False so we don't leak this helper thread.
            # If we made it True, an async cancellation here would detach us
            # from the helper thread and allow the thread to "run to completion"--
            # but if no more values are ever enqueued, and the queue is never closed,
            # completion would never happen and it would hang around forever.
            cancellable=False,
        )

    async def get_async_until_closed(self) -> AsyncIterable[_T]:
        """Like `get_until_closed()`, except yield to the event loop while waiting.

        Example:
            async for value in queue.get_async_until_closed():
                print(value)

        Warning:
            While the ``async for`` is waiting for a new value,
            it won't be interrupted by an async cancellation.
            The proper way to interrupt a waiting `get_async_until_closed()`
            is to close the queue, just like you have to do with `get()`.
        """
        while True:
            try:
                yield await self.get_async()
            except QueueClosed:
                break

    def done_putting(self) -> None:
        """Close the queue, i.e. signal that no more values will be `put()`.

        You normally *must* close the queue eventually
        to inform consumers that they can stop waiting for new values.
        Forgetting to do this can leave them waiting forever,
        leaking tasks or threads or causing deadlocks.

        Consider using a ``with`` block instead. See `__enter__()`.

        Raises:
            QueueClosed: If the queue is already closed.
        """
        with self._condition:
            if self._is_closed:
                raise QueueClosed("Can't close when queue is already closed.")
            else:
                self._is_closed = True
                self._condition.notify_all()

    def __enter__(self) -> ThreadAsyncQueue[_T]:
        """Use the queue as a context manager, closing the queue upon exit.

        Example:
            This:

                with queue:
                    do_stuff()

            Is equivalent to:

                try:
                    do_stuff()
                finally:
                    queue.done_putting()
        """
        return self

    def __exit__(self, exc_type: object, exc_val: object, exc_tb: object) -> None:
        """See `__enter__()`."""
        self.done_putting()


class QueueClosed(Exception):
    """See `ThreadAsyncQueue.done_putting()`."""

    pass



from __future__ import annotations
from enum import Enum
from dataclasses import dataclass
from typing import (
    Dict,
    NamedTuple,
    Callable,
    Any,
    Tuple,
    Awaitable,
    Union,
    cast,
    TYPE_CHECKING,
)
from typing_extensions import TypedDict
from pathlib import Path

from opentrons.drivers.rpi_drivers.types import USBPort

if TYPE_CHECKING:
    from opentrons_shared_data.module.dev_types import (
        ThermocyclerModuleType,
        MagneticModuleType,
        TemperatureModuleType,
        HeaterShakerModuleType,
        MagneticBlockType,
    )


class ThermocyclerStepBase(TypedDict):
    temperature: float


class ThermocyclerStep(ThermocyclerStepBase, total=False):
    hold_time_seconds: float
    hold_time_minutes: float


UploadFunction = Callable[[str, str, Dict[str, Any]], Awaitable[Tuple[bool, str]]]


class LiveData(TypedDict):
    status: str
    data: Dict[str, Union[float, str, bool, None]]


class ModuleType(str, Enum):
    THERMOCYCLER: ThermocyclerModuleType = "thermocyclerModuleType"
    TEMPERATURE: TemperatureModuleType = "temperatureModuleType"
    MAGNETIC: MagneticModuleType = "magneticModuleType"
    HEATER_SHAKER: HeaterShakerModuleType = "heaterShakerModuleType"
    MAGNETIC_BLOCK: MagneticBlockType = "magneticBlockType"

    @classmethod
    def from_model(cls, model: ModuleModel) -> ModuleType:
        if isinstance(model, MagneticModuleModel):
            return cls.MAGNETIC
        if isinstance(model, TemperatureModuleModel):
            return cls.TEMPERATURE
        if isinstance(model, ThermocyclerModuleModel):
            return cls.THERMOCYCLER
        if isinstance(model, HeaterShakerModuleModel):
            return cls.HEATER_SHAKER
        if isinstance(model, MagneticBlockModel):
            return cls.MAGNETIC_BLOCK


class MagneticModuleModel(str, Enum):
    MAGNETIC_V1: str = "magneticModuleV1"
    MAGNETIC_V2: str = "magneticModuleV2"


class TemperatureModuleModel(str, Enum):
    TEMPERATURE_V1: str = "temperatureModuleV1"
    TEMPERATURE_V2: str = "temperatureModuleV2"


class ThermocyclerModuleModel(str, Enum):
    THERMOCYCLER_V1: str = "thermocyclerModuleV1"
    THERMOCYCLER_V2: str = "thermocyclerModuleV2"


class HeaterShakerModuleModel(str, Enum):
    HEATER_SHAKER_V1: str = "heaterShakerModuleV1"


class MagneticBlockModel(str, Enum):
    MAGNETIC_BLOCK_V1: str = "magneticBlockV1"


def module_model_from_string(model_string: str) -> ModuleModel:
    for model_enum in {
        MagneticModuleModel,
        TemperatureModuleModel,
        ThermocyclerModuleModel,
        HeaterShakerModuleModel,
        MagneticBlockModel,
    }:
        try:
            return cast(ModuleModel, model_enum(model_string))
        except ValueError:
            pass
    raise ValueError(f"No such module model {model_string}")


@dataclass
class ModuleAtPort:
    port: str
    name: str
    usb_port: USBPort = USBPort(name="", port_number=0)


class BundledFirmware(NamedTuple):
    """Represents a versioned firmware file, generally bundled into the fs"""

    version: str
    path: Path

    def __repr__(self) -> str:
        return f"<BundledFirmware {self.version}, path={self.path}>"


class UpdateError(RuntimeError):
    pass


class ModuleInfo(NamedTuple):
    model: str  # A module model such as "magneticModuleV2"
    fw_version: str  # The version of the firmware
    hw_revision: str  # the revision of the hardware
    serial: str  # the serial number


# TODO(mc, 2022-01-18): replace with enum
ModuleModel = Union[
    MagneticModuleModel,
    TemperatureModuleModel,
    ThermocyclerModuleModel,
    HeaterShakerModuleModel,
    MagneticBlockModel,
]


class MagneticStatus(str, Enum):
    ENGAGED = "engaged"
    DISENGAGED = "disengaged"


class TemperatureStatus(str, Enum):
    HOLDING = "holding at target"
    COOLING = "cooling"
    HEATING = "heating"
    IDLE = "idle"
    ERROR = "error"


class SpeedStatus(str, Enum):
    HOLDING = "holding at target"
    ACCELERATING = "speeding up"
    DECELERATING = "slowing down"
    IDLE = "idle"
    ERROR = "error"


class HeaterShakerStatus(str, Enum):
    IDLE = "idle"
    RUNNING = "running"
    ERROR = "error"



"""Getters for specific adjacent slots."""
from dataclasses import dataclass
from typing import Optional, List, Dict, Union

from opentrons_shared_data.robot.dev_types import RobotType

from opentrons.types import DeckSlotName, StagingSlotName


def get_north_slot(slot: int) -> Optional[int]:
    """Get slot north of the given slot."""
    if slot in [10, 11, 12]:
        return None
    else:
        return slot + 3


def get_south_slot(slot: int) -> Optional[int]:
    """Get slot south of the given slot."""
    if slot in [1, 2, 3]:
        return None
    else:
        return slot - 3


def get_east_slot(slot: int) -> Optional[int]:
    """Get slot east of the given slot."""
    if slot in [3, 6, 9, 12]:
        return None
    else:
        return slot + 1


def get_west_slot(slot: int) -> Optional[int]:
    """Get slot west of the given slot."""
    if slot in [1, 4, 7, 10]:
        return None
    else:
        return slot - 1


def get_north_west_slot(slot: int) -> Optional[int]:
    """Get the slot that's north-west of the given slot."""
    if slot in [1, 4, 7, 10, 11, 12]:
        return None
    else:
        north_slot = get_north_slot(slot)
        return north_slot - 1 if north_slot else None


def get_north_east_slot(slot: int) -> Optional[int]:
    """Get the slot that's north-east of the given slot."""
    if slot in [3, 6, 9, 10, 11, 12]:
        return None
    else:
        north_slot = get_north_slot(slot)
        return north_slot + 1 if north_slot else None


def get_south_west_slot(slot: int) -> Optional[int]:
    """Get the slot that's south-west of the given slot."""
    if slot in [1, 2, 3, 4, 7, 10]:
        return None
    else:
        south_slot = get_south_slot(slot)
        return south_slot - 1 if south_slot else None


def get_south_east_slot(slot: int) -> Optional[int]:
    """Get the slot that's south-east of the given slot."""
    if slot in [1, 2, 3, 6, 9, 12]:
        return None
    else:
        south_slot = get_south_slot(slot)
        return south_slot + 1 if south_slot else None


@dataclass
class _MixedTypeSlots:
    regular_slots: List[DeckSlotName]
    staging_slots: List[StagingSlotName]


def get_surrounding_slots(slot: int, robot_type: RobotType) -> _MixedTypeSlots:
    """Get all the surrounding slots, i.e., adjacent slots as well as corner slots."""
    corner_slots: List[Union[int, None]] = [
        get_north_east_slot(slot),
        get_north_west_slot(slot),
        get_south_east_slot(slot),
        get_south_west_slot(slot),
    ]

    surrounding_regular_slots_int = get_adjacent_slots(slot) + [
        maybe_slot for maybe_slot in corner_slots if maybe_slot is not None
    ]
    surrounding_regular_slots = [
        DeckSlotName.from_primitive(slot_int).to_equivalent_for_robot_type(robot_type)
        for slot_int in surrounding_regular_slots_int
    ]
    surrounding_staging_slots = _SURROUNDING_STAGING_SLOTS_MAP.get(
        DeckSlotName.from_primitive(slot).to_equivalent_for_robot_type(robot_type), []
    )
    return _MixedTypeSlots(
        regular_slots=surrounding_regular_slots, staging_slots=surrounding_staging_slots
    )


_WEST_OF_STAGING_SLOT_MAP: Dict[StagingSlotName, DeckSlotName] = {
    StagingSlotName.SLOT_A4: DeckSlotName.SLOT_A3,
    StagingSlotName.SLOT_B4: DeckSlotName.SLOT_B3,
    StagingSlotName.SLOT_C4: DeckSlotName.SLOT_C3,
    StagingSlotName.SLOT_D4: DeckSlotName.SLOT_D3,
}

_EAST_OF_FLEX_COLUMN_3_MAP: Dict[DeckSlotName, StagingSlotName] = {
    deck_slot: staging_slot
    for staging_slot, deck_slot in _WEST_OF_STAGING_SLOT_MAP.items()
}


_SURROUNDING_STAGING_SLOTS_MAP: Dict[DeckSlotName, List[StagingSlotName]] = {
    DeckSlotName.SLOT_D3: [StagingSlotName.SLOT_C4, StagingSlotName.SLOT_D4],
    DeckSlotName.SLOT_C3: [
        StagingSlotName.SLOT_B4,
        StagingSlotName.SLOT_C4,
        StagingSlotName.SLOT_D4,
    ],
    DeckSlotName.SLOT_B3: [
        StagingSlotName.SLOT_A4,
        StagingSlotName.SLOT_B4,
        StagingSlotName.SLOT_C4,
    ],
    DeckSlotName.SLOT_A3: [StagingSlotName.SLOT_A4, StagingSlotName.SLOT_B4],
}


def get_west_of_staging_slot(staging_slot: StagingSlotName) -> DeckSlotName:
    """Get slot west of a staging slot."""
    return _WEST_OF_STAGING_SLOT_MAP[staging_slot]


def get_adjacent_staging_slot(deck_slot: DeckSlotName) -> Optional[StagingSlotName]:
    """Get the adjacent staging slot if the deck slot is in the third column."""
    return _EAST_OF_FLEX_COLUMN_3_MAP.get(deck_slot)


def get_surrounding_staging_slots(deck_slot: DeckSlotName) -> List[StagingSlotName]:
    """Get the staging slots surrounding the given deck slot."""
    return _SURROUNDING_STAGING_SLOTS_MAP.get(deck_slot, [])


def get_east_west_slots(slot: int) -> List[int]:
    """Get slots east & west of the given slot."""
    east = get_east_slot(slot)
    west = get_west_slot(slot)
    return [maybe_slot for maybe_slot in [east, west] if maybe_slot is not None]


def get_north_south_slots(slot: int) -> List[int]:
    """Get slots north & south of the given slot."""
    north = get_north_slot(slot)
    south = get_south_slot(slot)
    return [maybe_slot for maybe_slot in [north, south] if maybe_slot is not None]


def get_adjacent_slots(slot: int) -> List[int]:
    """Get slots on the east, west, north and south of the given slot."""
    return get_east_west_slots(slot) + get_north_south_slots(slot)



"""Protocol engine plugin interface."""
from __future__ import annotations
from abc import ABC, abstractmethod
from typing import List
from typing_extensions import final

from .actions import Action, ActionDispatcher, ActionHandler
from .state import StateView


class AbstractPlugin(ActionHandler, ABC):
    """An ProtocolEngine plugin to customize engine behavior.

    A plugin may customize behavior in one of two ways:

    1. It can react to actions as they flow through the action pipeline,
       before they hit the StateStore.
    2. It can dispatch new actions into the pipeline.
    """

    @final
    @property
    def state(self) -> StateView:
        """Get the current ProtocolEngine state."""
        return self._state

    @final
    def dispatch(self, action: Action) -> None:
        """Dispatch an action into the action pipeline.

        Arguments:
            action: A new ProtocolEngine action to send into the pipeline.
                This action will flow through all plugins, including
                this one, so be careful to avoid infinite loops. In general,
                do not dispatch an action your plugin will react to.
        """
        return self._action_dispatcher.dispatch(action)

    def setup(self) -> None:
        """Run any necessary setup steps prior to plugin usage."""
        ...

    async def teardown(self) -> None:
        """Run any necessary teardown steps once the engine is stopped."""
        ...

    @abstractmethod
    def handle_action(self, action: Action) -> None:
        """React to an action going through the pipeline.

        When reacting to an action, `self.state` will not yet
        reflect the change represented by the action, because
        plugins receive actions before the StateStore.

        Arguments:
            action: An action that has been dispatched into the pipeline
                that this plugin may react to.
        """
        ...

    # NOTE: using a "protected" method allows mypy to infer the types
    # of private properties `_state` and `_action_dispatcher` without
    # declaring them on the class
    @final
    def _configure(
        self,
        state: StateView,
        action_dispatcher: ActionDispatcher,
    ) -> None:
        """Insert a StateView and ActionDispatcher into the plugin.

        This is a protected method that should only be called internally
        by the ProtocolEngine during plugin setup.
        """
        self._state = state
        self._action_dispatcher = action_dispatcher


class PluginStarter:
    """Configure, setup, and tear down engine plugins."""

    def __init__(
        self,
        state: StateView,
        action_dispatcher: ActionDispatcher,
    ) -> None:
        """Create a PluginStarter interface with access to its dependencies."""
        self._state = state
        self._action_dispatcher = action_dispatcher
        self._plugins: List[AbstractPlugin] = []

    def start(self, plugin: AbstractPlugin) -> None:
        """Configure a given plugin and add it to the dispatch pipeline."""
        plugin._configure(state=self._state, action_dispatcher=self._action_dispatcher)
        self._plugins.append(plugin)
        self._action_dispatcher.add_handler(plugin)
        plugin.setup()

    async def stop(self) -> None:
        """Stop any configured plugins."""
        for p in self._plugins:
            await p.teardown()

        self._plugins = []



"""Models and implementation for the calibrateGripper command."""

from enum import Enum
from typing import Optional, Type
from typing_extensions import Literal

from pydantic import BaseModel, Field

from opentrons.types import Point
from opentrons.hardware_control import HardwareControlAPI
from opentrons.hardware_control import ot3_calibration
from opentrons.hardware_control.types import OT3Mount, GripperProbe as HWAPIGripperProbe
from opentrons.hardware_control.instruments.ot3.instrument_calibration import (
    GripperCalibrationOffset,
)
from opentrons.protocol_engine.commands.command import (
    AbstractCommandImpl,
    BaseCommand,
    BaseCommandCreate,
)
from opentrons.protocol_engine.types import Vec3f
from opentrons.protocol_engine.resources import ensure_ot3_hardware


CalibrateGripperCommandType = Literal["calibration/calibrateGripper"]


class CalibrateGripperParamsJaw(Enum):  # noqa: D101
    FRONT = "front"
    REAR = "rear"


class CalibrateGripperParams(BaseModel):
    """Parameters for a `calibrateGripper` command."""

    jaw: CalibrateGripperParamsJaw = Field(
        ...,
        description=(
            "Which of the gripper's jaws to use to measure its offset."
            " The robot will assume that a human operator has already attached"
            " the capacitive probe to the jaw and none is attached to the other jaw."
        ),
    )

    otherJawOffset: Optional[Vec3f] = Field(
        None,
        description=(
            "If an offset for the other probe is already found, then specifying it here"
            " will enable the CalibrateGripper command to complete the calibration"
            " process by calculating the total offset and saving it to disk."
            " If this param is not specified then the command will only find and return"
            " the offset for the specified probe."
        ),
    )


class CalibrateGripperResult(BaseModel):
    """The result of a successful `calibrateGripper` command."""

    jawOffset: Vec3f = Field(
        ...,
        description=(
            "The offset from the probe's nominal position"
            " to its actual measured position."
        ),
    )

    savedCalibration: Optional[GripperCalibrationOffset] = Field(
        None,
        description=(
            "Gripper calibration result data, when `otherJawOffset` is provided."
        ),
    )


class CalibrateGripperImplementation(
    AbstractCommandImpl[CalibrateGripperParams, CalibrateGripperResult]
):
    """The implementation of a `calibrateGripper` command."""

    def __init__(
        self,
        hardware_api: HardwareControlAPI,
        *args: object,
        **kwargs: object,
    ) -> None:
        self._hardware_api = hardware_api

    async def execute(self, params: CalibrateGripperParams) -> CalibrateGripperResult:
        """Execute a `calibrateGripper` command.

        1. Move from the current location to the calibration area on the deck.
        2. Using the given probe, sense the gripper's offset from its ideal position.
        3. Return the gripper's offset from its ideal position.
        """
        ot3_hardware_api = ensure_ot3_hardware(self._hardware_api)

        probe_offset = await ot3_calibration.calibrate_gripper_jaw(
            hcapi=ot3_hardware_api, probe=self._convert_to_hw_api_probe(params.jaw)
        )
        other_probe_offset = params.otherJawOffset
        calibration_data: Optional[GripperCalibrationOffset] = None
        if other_probe_offset is not None:
            total_offset = 0.5 * (
                probe_offset
                + Point(
                    x=other_probe_offset.x,
                    y=other_probe_offset.y,
                    z=other_probe_offset.z,
                )
            )
            result = await ot3_hardware_api.save_instrument_offset(
                mount=OT3Mount.GRIPPER, delta=total_offset
            )
            assert isinstance(result, GripperCalibrationOffset), (
                f"Expected result to be GripperCalibrationOffset type, "
                f"but received {type(result)}"
            )
            calibration_data = result

        return CalibrateGripperResult.construct(
            jawOffset=Vec3f.construct(
                x=probe_offset.x, y=probe_offset.y, z=probe_offset.z
            ),
            savedCalibration=calibration_data,
        )

    @staticmethod
    def _convert_to_hw_api_probe(
        probe_from_params: CalibrateGripperParamsJaw,
    ) -> HWAPIGripperProbe:
        if probe_from_params is CalibrateGripperParamsJaw.FRONT:
            return HWAPIGripperProbe.FRONT
        elif probe_from_params is CalibrateGripperParamsJaw.REAR:
            return HWAPIGripperProbe.REAR


class CalibrateGripper(BaseCommand[CalibrateGripperParams, CalibrateGripperResult]):
    """A `calibrateGripper` command."""

    commandType: CalibrateGripperCommandType = "calibration/calibrateGripper"
    params: CalibrateGripperParams
    result: Optional[CalibrateGripperResult]

    _ImplementationCls: Type[
        CalibrateGripperImplementation
    ] = CalibrateGripperImplementation


class CalibrateGripperCreate(BaseCommandCreate[CalibrateGripperParams]):
    """A request to create a `calibrateGripper` command."""

    commandType: CalibrateGripperCommandType = "calibration/calibrateGripper"
    params: CalibrateGripperParams

    _CommandCls: Type[CalibrateGripper] = CalibrateGripper



"""Contains configuration related classes for Protocol API."""


class Clearances:
    def __init__(self, default_aspirate: float, default_dispense: float) -> None:
        self._aspirate = default_aspirate
        self._dispense = default_dispense

    @property
    def aspirate(self) -> float:
        return self._aspirate

    @aspirate.setter
    def aspirate(self, new_val: float) -> None:
        self._aspirate = float(new_val)

    @property
    def dispense(self) -> float:
        return self._dispense

    @dispense.setter
    def dispense(self, new_val: float) -> None:
        self._dispense = float(new_val)



from typing import List, Tuple

from opentrons_shared_data.pipette.pipette_definition import PipetteFunctionKeyType

PIPETTING_FUNCTION_FALLBACK_VERSION: PipetteFunctionKeyType = "1"
PIPETTING_FUNCTION_LATEST_VERSION: PipetteFunctionKeyType = "2"


def piecewise_volume_conversion(
    ul: float, sequence: List[Tuple[float, float, float]]
) -> float:
    """
    Takes a volume in microliters and a sequence representing a piecewise
    function for the slope and y-intercept of a ul/mm function, where each
    sub-list in the sequence contains:

      - the max volume for the piece of the function (minimum implied from the
        max of the previous item or 0
      - the slope of the segment
      - the y-intercept of the segment

    :return: the ul/mm value for the specified volume
    """
    # pick the first item from the seq for which the target is less than
    # the bracketing element
    for x in sequence:
        if ul <= x[0]:
            # use that element to calculate the movement distance in mm
            return x[1] * ul + x[2]

    # Compatibility with previous implementation of search.
    #  list(filter(lambda x: ul <= x[0], sequence))[0]
    raise IndexError()



"""Shared code for managing pipette configuration and storage."""
from dataclasses import dataclass
import logging
from typing import (
    Callable,
    Dict,
    Generic,
    Optional,
    Tuple,
    Any,
    cast,
    List,
    Sequence,
    Iterator,
    TypeVar,
    overload,
)
import numpy

from opentrons_shared_data.errors.exceptions import (
    UnexpectedTipRemovalError,
    UnexpectedTipAttachError,
)
from opentrons_shared_data.pipette.dev_types import UlPerMmAction
from opentrons_shared_data.pipette.types import Quirks
from opentrons_shared_data.errors.exceptions import CommandPreconditionViolated

from opentrons import types as top_types
from opentrons.hardware_control.types import (
    CriticalPoint,
    HardwareAction,
    Axis,
    OT3Mount,
)
from opentrons.hardware_control.constants import (
    SHAKE_OFF_TIPS_SPEED,
    SHAKE_OFF_TIPS_PICKUP_DISTANCE,
    DROP_TIP_RELEASE_DISTANCE,
    SHAKE_OFF_TIPS_DROP_DISTANCE,
)

from opentrons.hardware_control.dev_types import PipetteDict
from .pipette import Pipette

# TODO both pipette handlers should be combined once the pipette configurations
# are unified AND we separate out the concept of changing pipette state versus static state

MountType = TypeVar("MountType", top_types.Mount, OT3Mount)

InstrumentsByMount = Dict[MountType, Optional[Pipette]]
PipetteHandlingData = Tuple[Pipette, MountType]

MOD_LOG = logging.getLogger(__name__)


@dataclass(frozen=True)
class LiquidActionSpec:
    axis: Axis
    volume: float
    plunger_distance: float
    speed: float
    instr: Pipette
    current: float


@dataclass(frozen=True)
class PickUpTipPressSpec:
    relative_down: top_types.Point
    relative_up: top_types.Point
    current: Dict[Axis, float]
    speed: float


@dataclass(frozen=True)
class PickUpTipSpec:
    plunger_prep_pos: float
    plunger_currents: Dict[Axis, float]
    presses: List[PickUpTipPressSpec]
    shake_off_list: List[Tuple[top_types.Point, Optional[float]]]
    retract_target: float


@dataclass(frozen=True)
class DropTipMove:
    target_position: float
    current: Dict[Axis, float]
    speed: Optional[float]
    home_after: bool = False
    home_after_safety_margin: float = 0
    home_axes: Sequence[Axis] = tuple()


@dataclass(frozen=True)
class DropTipSpec:
    drop_moves: List[DropTipMove]
    shake_moves: List[Tuple[top_types.Point, Optional[float]]]
    ending_current: Dict[Axis, float]


class PipetteHandlerProvider(Generic[MountType]):
    IHP_LOG = MOD_LOG.getChild("InstrumentHandler")

    def __init__(self, attached_instruments: InstrumentsByMount[MountType]):
        assert attached_instruments
        self._attached_instruments: InstrumentsByMount[MountType] = attached_instruments
        self._ihp_log = PipetteHandlerProvider.IHP_LOG.getChild(str(id(self)))

    def reset_instrument(self, mount: Optional[MountType] = None) -> None:
        """
        Reset the internal state of a pipette by its mount, without doing
        any lower level reconfiguration. This is useful to make sure that no
        settings changes from a protocol persist.

        :param mount: If specified, reset that mount. If not specified,
                      reset both
        """

        def _reset(m: MountType) -> None:
            if isinstance(m, top_types.Mount) and m not in top_types.Mount.ot2_mounts():
                self._ihp_log.warning(
                    "Received a non OT2 mount for resetting. Skipping"
                )
                return
            self._ihp_log.info(f"Resetting configuration for {m}")
            p = self._attached_instruments[m]
            if not p:
                return
            if isinstance(m, OT3Mount):
                # This is to satisfy lint. Code will be cleaner once
                # we can combine the pipette handler for OT2 and OT3
                # pipettes again.
                p.reset_pipette_offset(m.to_mount(), to_default=False)
            else:
                p.reset_pipette_offset(m, to_default=False)
            p.reload_configurations()
            p.reset_state()

        if not mount:
            for m in type(list(self._attached_instruments.keys())[0]):
                _reset(m)
        else:
            _reset(mount)

    def reset_instrument_offset(self, mount: MountType, to_default: bool) -> None:
        """
        Temporarily reset the pipette offset to default values.
        :param mount: Modify the given mount.
        """
        if isinstance(mount, OT3Mount):
            # This is to satisfy lint. Code will be cleaner once
            # we can combine the pipette handler for OT2 and OT3
            # pipettes again.
            pipette = self.get_pipette(mount)
            pipette.reset_pipette_offset(mount.to_mount(), to_default)
        else:
            pipette = self.get_pipette(mount)
            pipette.reset_pipette_offset(mount, to_default)

    def save_instrument_offset(self, mount: MountType, delta: top_types.Point) -> None:
        """
        Save a new instrument offset the pipette offset to a particular value.
        :param mount: Modify the given mount.
        :param delta: The offset to set for the pipette.
        """
        if isinstance(mount, OT3Mount):
            # This is to satisfy lint. Code will be cleaner once
            # we can combine the pipette handler for OT2 and OT3
            # pipettes again.
            pipette = self.get_pipette(mount)
            pipette.save_pipette_offset(mount.to_mount(), delta)
        else:
            pipette = self.get_pipette(mount)
            pipette.save_pipette_offset(mount, delta)

    # TODO(mc, 2022-01-11): change returned map value type to `Optional[PipetteDict]`
    # instead of potentially returning an empty dict
    def get_attached_instruments(self) -> Dict[MountType, PipetteDict]:
        """Get the status dicts of the cached attached instruments.

        Also available as :py:meth:`get_attached_instruments`.

        This returns a dictified version of the
        :py:class:`hardware_control.instruments.pipette.Pipette` as a dict keyed by
        the :py:class:`top_types.Mount` to which the pipette is attached.
        If no pipette is attached on a given mount, the mount key will
        still be present but will have the value ``None``.

        Note that this is only a query of a cached value; to actively scan
        for changes, use :py:meth:`cache_instruments`. This process deactivates
        the motors and should be used sparingly.
        """
        return {
            m: self.get_attached_instrument(m)
            for m in self._attached_instruments.keys()
        }

    # TODO(mc, 2022-01-11): change return type to `Optional[PipetteDict]` instead
    # of potentially returning an empty dict
    def get_attached_instrument(self, mount: MountType) -> PipetteDict:
        instr = self._attached_instruments[mount]
        result: Dict[str, Any] = {}
        if instr:
            configs = [
                "name",
                "aspirate_flow_rate",
                "dispense_flow_rate",
                "pipette_id",
                "current_volume",
                "display_name",
                "tip_length",
                "model",
                "blow_out_flow_rate",
                "working_volume",
                "tip_overlap",
                "available_volume",
                "return_tip_height",
                "default_aspirate_flow_rates",
                "default_blow_out_flow_rates",
                "default_dispense_flow_rates",
                "back_compat_names",
                "supported_tips",
            ]

            instr_dict = instr.as_dict()
            # TODO (spp, 2021-08-27): Revisit this logic. Why do we need to build
            #  this dict newly every time? Any why only a few items are being updated?
            for key in configs:
                result[key] = instr_dict[key]
            result["current_nozzle_map"] = instr.nozzle_manager.current_configuration
            result["min_volume"] = instr.liquid_class.min_volume
            result["max_volume"] = instr.liquid_class.max_volume
            result["channels"] = instr.channels
            result["has_tip"] = instr.has_tip
            result["tip_length"] = instr.current_tip_length
            result["aspirate_speed"] = self.plunger_speed(
                instr, instr.aspirate_flow_rate, "aspirate"
            )
            result["dispense_speed"] = self.plunger_speed(
                instr, instr.dispense_flow_rate, "dispense"
            )
            result["blow_out_speed"] = self.plunger_speed(
                instr, instr.blow_out_flow_rate, "dispense"
            )
            result["ready_to_aspirate"] = instr.ready_to_aspirate
            result["default_blow_out_speeds"] = {
                alvl: self.plunger_speed(instr, fr, "blowout")
                for alvl, fr in instr.blow_out_flow_rates_lookup.items()
            }

            result["default_dispense_speeds"] = {
                alvl: self.plunger_speed(instr, fr, "dispense")
                for alvl, fr in instr.dispense_flow_rates_lookup.items()
            }
            result["default_aspirate_speeds"] = {
                alvl: self.plunger_speed(instr, fr, "aspirate")
                for alvl, fr in instr.aspirate_flow_rates_lookup.items()
            }
            result[
                "pipette_bounding_box_offsets"
            ] = instr.config.pipette_bounding_box_offsets
        return cast(PipetteDict, result)

    @property
    def attached_instruments(self) -> Dict[MountType, PipetteDict]:
        return self.get_attached_instruments()

    @property
    def attached_pipettes(self) -> Dict[MountType, PipetteDict]:
        return self.get_attached_instruments()

    @property
    def get_attached_pipettes(self) -> Dict[MountType, PipetteDict]:
        return self.get_attached_instruments()

    @property
    def hardware_instruments(self) -> InstrumentsByMount[MountType]:
        """Do not write new code that uses this."""
        return self._attached_instruments

    def set_current_tiprack_diameter(
        self, mount: MountType, tiprack_diameter: float
    ) -> None:
        instr = self.get_pipette(mount)
        self._ihp_log.info(
            "Updating tip rack diameter on pipette mount: "
            f"{mount}, tip diameter: {tiprack_diameter} mm"
        )
        instr.current_tiprack_diameter = tiprack_diameter

    def set_working_volume(self, mount: MountType, tip_volume: float) -> None:
        instr = self.get_pipette(mount)
        if not instr:
            raise top_types.PipetteNotAttachedError(
                "No pipette attached to {} mount".format(mount.name)
            )
        self._ihp_log.info(
            "Updating working volume on pipette mount:"
            f"{mount}, tip volume: {tip_volume} ul"
        )
        instr.working_volume = tip_volume

    def calibrate_plunger(
        self,
        mount: MountType,
        top: Optional[float] = None,
        bottom: Optional[float] = None,
        blow_out: Optional[float] = None,
        drop_tip: Optional[float] = None,
    ) -> None:
        """
        Set calibration values for the pipette plunger.
        This can be called multiple times as the user sets each value,
        or you can set them all at once.
        :param top: Touching but not engaging the plunger.
        :param bottom: Must be above the pipette's physical hard-stop, while
        still leaving enough room for 'blow_out'
        :param blow_out: Plunger is pushed down enough to expel all liquids.
        :param drop_tip: Position that causes the tip to be released from the
        pipette
        """
        instr = self.get_pipette(mount)
        pos_dict: Dict[str, float] = {
            "top": instr.plunger_positions.top,
            "bottom": instr.plunger_positions.bottom,
            "blow_out": instr.plunger_positions.blow_out,
            "drop_tip": instr.plunger_positions.drop_tip,
        }
        if top is not None:
            pos_dict["top"] = top
        if bottom is not None:
            pos_dict["bottom"] = bottom
        if blow_out is not None:
            pos_dict["blow_out"] = blow_out
        if drop_tip is not None:
            pos_dict["drop_tip"] = drop_tip
        instr.update_config_item(pos_dict)

    def set_flow_rate(
        self,
        mount: MountType,
        aspirate: Optional[float] = None,
        dispense: Optional[float] = None,
        blow_out: Optional[float] = None,
    ) -> None:
        this_pipette = self.get_pipette(mount)
        if aspirate:
            this_pipette.aspirate_flow_rate = aspirate
        if dispense:
            this_pipette.dispense_flow_rate = dispense
        if blow_out:
            this_pipette.blow_out_flow_rate = blow_out

    def set_pipette_speed(
        self,
        mount: MountType,
        aspirate: Optional[float] = None,
        dispense: Optional[float] = None,
        blow_out: Optional[float] = None,
    ) -> None:
        this_pipette = self.get_pipette(mount)
        if aspirate:
            this_pipette.aspirate_flow_rate = self.plunger_flowrate(
                this_pipette, aspirate, "aspirate"
            )
        if dispense:
            this_pipette.dispense_flow_rate = self.plunger_flowrate(
                this_pipette, dispense, "dispense"
            )
        if blow_out:
            this_pipette.blow_out_flow_rate = self.plunger_flowrate(
                this_pipette, blow_out, "dispense"
            )

    def instrument_max_height(
        self,
        mount: MountType,
        retract_distance: float,
        critical_point: Optional[CriticalPoint],
    ) -> float:
        """Return max achievable height of the attached instrument
        based on the current critical point
        """
        pip = self.get_pipette(mount)
        cp = self.critical_point_for(mount, critical_point)

        max_height = (
            pip.config.mount_configurations.homePosition - retract_distance + cp.z
        )

        return max_height

    async def reset(self) -> None:
        self._attached_instruments = {
            k: None for k in self._attached_instruments.keys()
        }

    async def update_nozzle_configuration(
        self,
        mount: MountType,
        back_left_nozzle: str,
        front_right_nozzle: str,
        starting_nozzle: Optional[str] = None,
    ) -> None:
        instr = self._attached_instruments[mount]
        if instr:
            instr.update_nozzle_configuration(
                back_left_nozzle, front_right_nozzle, starting_nozzle
            )

    async def reset_nozzle_configuration(self, mount: MountType) -> None:
        instr = self._attached_instruments[mount]
        if instr:
            instr.reset_nozzle_configuration()

    async def add_tip(self, mount: MountType, tip_length: float) -> None:
        instr = self._attached_instruments[mount]
        attached = self.attached_instruments
        instr_dict = attached[mount]
        if instr and not instr.has_tip:
            instr.add_tip(tip_length=tip_length)
            # TODO (spp, 2021-08-27): These items are being updated in a local copy
            #  of the PipetteDict, which gets thrown away. Fix this.
            instr_dict["has_tip"] = True
            instr_dict["tip_length"] = tip_length
        else:
            self._ihp_log.warning(
                f"attach tip called while tip already attached to {instr}"
            )

    async def remove_tip(self, mount: MountType) -> None:
        instr = self._attached_instruments[mount]
        attached = self.attached_instruments
        instr_dict = attached[mount]
        if instr and instr.has_tip:
            instr.remove_tip()
            # TODO (spp, 2021-08-27): These items are being updated in a local copy
            #  of the PipetteDict, which gets thrown away. Fix this.
            instr_dict["has_tip"] = False
            instr_dict["tip_length"] = 0.0
        else:
            self._ihp_log.warning("detach tip called with no tip")

    def critical_point_for(
        self, mount: MountType, cp_override: Optional[CriticalPoint] = None
    ) -> top_types.Point:
        """Return the current critical point of the specified mount.

        The mount's critical point is the position of the mount itself, if no
        pipette is attached, or the pipette's critical point (which depends on
        tip status).

        If `cp_override` is specified, and that critical point actually exists,
        it will be used instead. Invalid `cp_override`s are ignored.
        """
        pip = self._attached_instruments[mount]
        if pip is not None and cp_override != CriticalPoint.MOUNT:
            return pip.critical_point(cp_override)
        else:
            # This offset is required because the motor driver coordinate system is
            # configured such that the end of a p300 single gen1's tip is 0.
            return top_types.Point(0, 0, 30)

    def ready_for_tip_action(
        self, target: Pipette, action: HardwareAction, mount: MountType
    ) -> None:
        if not target.has_tip:
            raise UnexpectedTipRemovalError(action.name, target.name, mount.name)
        if (
            action == HardwareAction.ASPIRATE
            and target.current_volume == 0
            and not target.ready_to_aspirate
        ):
            raise RuntimeError("Pipette not ready to aspirate")
        self._ihp_log.debug(f"{action} on {target.name}")

    def plunger_position(
        self, instr: Pipette, ul: float, action: "UlPerMmAction"
    ) -> float:
        mm = ul / instr.ul_per_mm(ul, action)
        position = mm + instr.plunger_positions.bottom
        return round(position, 6)

    def plunger_speed(
        self, instr: Pipette, ul_per_s: float, action: "UlPerMmAction"
    ) -> float:
        mm_per_s = ul_per_s / instr.ul_per_mm(instr.liquid_class.max_volume, action)
        return round(mm_per_s, 6)

    def plunger_flowrate(
        self, instr: Pipette, mm_per_s: float, action: "UlPerMmAction"
    ) -> float:
        ul_per_s = mm_per_s * instr.ul_per_mm(instr.liquid_class.max_volume, action)
        return round(ul_per_s, 6)

    @overload
    def plan_check_aspirate(
        self, mount: top_types.Mount, volume: Optional[float], rate: float
    ) -> Optional[LiquidActionSpec]:
        ...

    @overload
    def plan_check_aspirate(
        self, mount: OT3Mount, volume: Optional[float], rate: float
    ) -> Optional[LiquidActionSpec]:
        ...

    # note on this type ignore: see motion_utilities
    def plan_check_aspirate(  # type: ignore[no-untyped-def]
        self,
        mount,
        volume,
        rate,
    ):
        """Check preconditions for aspirate, parse args, and calculate positions.

        While the mechanics of issuing an aspirate move itself are left to child
        classes, determining things like aspiration volume from the allowed argument
        types is invariant between machines, and this method gathers that functionality.

        Coalesce
        - Optional volumes

        Check
        - Aspiration volumes compared to max and remaining

        Calculate
        - Plunger distances (possibly calling an overridden plunger_volume)
        """
        instrument = self.get_pipette(mount)
        self.ready_for_tip_action(instrument, HardwareAction.ASPIRATE, mount)
        if volume is None:
            self._ihp_log.debug(
                "No aspirate volume defined. Aspirating up to "
                "max_volume for the pipette"
            )
            asp_vol = instrument.available_volume
        else:
            asp_vol = volume

        if asp_vol == 0:
            return None

        assert instrument.ok_to_add_volume(
            asp_vol
        ), "Cannot aspirate more than pipette max volume"

        dist = self.plunger_position(
            instrument, instrument.current_volume + asp_vol, "aspirate"
        )
        speed = self.plunger_speed(
            instrument, instrument.aspirate_flow_rate * rate, "aspirate"
        )
        if isinstance(mount, OT3Mount):
            return LiquidActionSpec(
                axis=Axis.of_main_tool_actuator(mount),
                volume=asp_vol,
                plunger_distance=dist,
                speed=speed,
                instr=instrument,
                current=instrument.plunger_motor_current.run,
            )
        else:
            return LiquidActionSpec(
                axis=Axis.of_plunger(mount),
                volume=asp_vol,
                plunger_distance=dist,
                speed=speed,
                instr=instrument,
                current=instrument.plunger_motor_current.run,
            )

    @overload
    def plan_check_dispense(
        self,
        mount: top_types.Mount,
        volume: Optional[float],
        rate: float,
        push_out: Optional[float],
    ) -> Optional[LiquidActionSpec]:
        ...

    @overload
    def plan_check_dispense(
        self,
        mount: OT3Mount,
        volume: Optional[float],
        rate: float,
        push_out: Optional[float],
    ) -> Optional[LiquidActionSpec]:
        ...

    def plan_check_dispense(  # type: ignore[no-untyped-def]
        self, mount, volume, rate, push_out
    ) -> Optional[LiquidActionSpec]:
        """Check preconditions for dispense, parse args, and calculate positions.

        While the mechanics of issuing a dispense move itself are left to child
        classes, determining things like dispense volume from the allowed argument
        types is invariant between machines, and this method gathers that functionality.

        Coalesce
        - Optional volumes

        Check
        - Dispense volumes compared to max and remaining

        Calculate
        - Plunger distances (possibly calling an overridden plunger_volume)
        """

        instrument = self.get_pipette(mount)
        self.ready_for_tip_action(instrument, HardwareAction.DISPENSE, mount)

        if volume is None:
            disp_vol = instrument.current_volume
            self._ihp_log.debug(
                "No dispense volume specified. Dispensing all "
                "remaining liquid ({}uL) from pipette".format(disp_vol)
            )
        else:
            disp_vol = volume

        # Ensure we don't dispense more than the current volume.
        #
        # This clamping is inconsistent with plan_check_aspirate(), which asserts
        # that its input is in bounds instead of clamping it. This is left to avoid
        # disturbing Python protocols with apiLevel <= 2.13. In newer Python protocols,
        # the Protocol Engine layer applies its own bounds checking.
        disp_vol = min(instrument.current_volume, disp_vol)

        if disp_vol == 0:
            return None

        is_full_dispense = numpy.isclose(instrument.current_volume - disp_vol, 0)

        if is_full_dispense:
            if push_out is None:
                push_out_ul = instrument.push_out_volume
            else:
                push_out_ul = push_out
        else:
            if push_out is not None and push_out != 0:
                raise CommandPreconditionViolated(
                    message="Cannot push_out on a dispense that does not leave the pipette empty",
                    detail={
                        "command": "dispense",
                        "remaining-volume": str(instrument.current_volume - disp_vol),
                    },
                )
            push_out_ul = 0

        push_out_dist_mm = push_out_ul / instrument.ul_per_mm(push_out_ul, "blowout")

        if not instrument.ok_to_push_out(push_out_dist_mm):
            raise CommandPreconditionViolated(
                message="Cannot push_out more than pipette max blowout volume.",
                detail={
                    "command": "dispense",
                },
            )

        dist = self.plunger_position(
            instrument, instrument.current_volume - disp_vol, "dispense"
        )
        speed = self.plunger_speed(
            instrument, instrument.dispense_flow_rate * rate, "dispense"
        )
        if isinstance(mount, top_types.Mount):
            return LiquidActionSpec(
                axis=Axis.of_plunger(mount),
                volume=disp_vol,
                plunger_distance=dist + push_out_dist_mm,
                speed=speed,
                instr=instrument,
                current=instrument.plunger_motor_current.run,
            )
        else:
            return LiquidActionSpec(
                axis=Axis.of_main_tool_actuator(mount),
                volume=disp_vol,
                plunger_distance=dist,
                speed=speed,
                instr=instrument,
                current=instrument.plunger_motor_current.run,
            )

    @overload
    def plan_check_blow_out(self, mount: top_types.Mount) -> LiquidActionSpec:
        ...

    @overload
    def plan_check_blow_out(self, mount: OT3Mount) -> LiquidActionSpec:
        ...

    def plan_check_blow_out(self, mount):  # type: ignore[no-untyped-def]
        """Check preconditions and calculate values for blowout."""
        instrument = self.get_pipette(mount)
        speed = self.plunger_speed(
            instrument, instrument.blow_out_flow_rate, "dispense"
        )

        if isinstance(mount, top_types.Mount):
            return LiquidActionSpec(
                axis=Axis.of_plunger(mount),
                volume=0,
                plunger_distance=instrument.plunger_positions.blow_out,
                speed=speed,
                instr=instrument,
                current=instrument.plunger_motor_current.run,
            )
        else:
            return LiquidActionSpec(
                axis=Axis.of_main_tool_actuator(mount),
                volume=0,
                plunger_distance=instrument.plunger_positions.blow_out,
                speed=speed,
                instr=instrument,
                current=instrument.plunger_motor_current.run,
            )

    @staticmethod
    def _build_pickup_shakes(
        instrument: Pipette,
    ) -> List[Tuple[top_types.Point, Optional[float]]]:
        def build_one_shake() -> List[Tuple[top_types.Point, Optional[float]]]:
            shake_dist = float(SHAKE_OFF_TIPS_PICKUP_DISTANCE)
            shake_speed = float(SHAKE_OFF_TIPS_SPEED)
            return [
                (top_types.Point(-shake_dist, 0, 0), shake_speed),  # left
                (top_types.Point(2 * shake_dist, 0, 0), shake_speed),  # right
                (top_types.Point(-shake_dist, 0, 0), shake_speed),  # center
                (top_types.Point(0, -shake_dist, 0), shake_speed),  # front
                (top_types.Point(0, 2 * shake_dist, 0), shake_speed),  # back
                (top_types.Point(0, -shake_dist, 0), shake_speed),  # center
                (top_types.Point(0, 0, DROP_TIP_RELEASE_DISTANCE), None),  # up
            ]

        if Quirks.pickupTipShake in instrument.config.quirks:
            return build_one_shake() + build_one_shake()
        else:
            return []

    @overload
    def plan_check_pick_up_tip(
        self,
        mount: top_types.Mount,
        tip_length: float,
        presses: Optional[int],
        increment: Optional[float],
    ) -> Tuple[PickUpTipSpec, Callable[[], None]]:
        ...

    @overload
    def plan_check_pick_up_tip(
        self,
        mount: OT3Mount,
        tip_length: float,
        presses: Optional[int],
        increment: Optional[float],
    ) -> Tuple[PickUpTipSpec, Callable[[], None]]:
        ...

    def plan_check_pick_up_tip(  # type: ignore[no-untyped-def]
        self,
        mount,
        tip_length,
        presses,
        increment,
    ):
        # Prechecks: ready for pickup tip and press/increment are valid
        instrument = self.get_pipette(mount)
        if instrument.has_tip:
            raise UnexpectedTipAttachError("pick_up_tip", instrument.name, mount.name)
        self._ihp_log.debug(f"Picking up tip on {mount.name}")

        if presses is None or presses < 0:
            checked_presses = instrument.pick_up_configurations.press_fit.presses
        else:
            checked_presses = presses

        if not increment or increment < 0:
            check_incr = instrument.pick_up_configurations.press_fit.increment
        else:
            check_incr = increment

        pick_up_speed = instrument.pick_up_configurations.press_fit.speed

        def build_presses() -> Iterator[Tuple[float, float]]:
            # Press the nozzle into the tip <presses> number of times,
            # moving further by <increment> mm after each press
            for i in range(checked_presses):
                # move nozzle down into the tip
                press_dist = (
                    -1.0 * instrument.pick_up_configurations.press_fit.distance
                    + -1.0 * check_incr * i
                )
                # move nozzle back up
                backup_dist = -press_dist
                yield (press_dist, backup_dist)

        def add_tip_to_instr() -> None:
            instrument.add_tip(tip_length=tip_length)
            instrument.set_current_volume(0)

        if isinstance(mount, top_types.Mount):
            return (
                PickUpTipSpec(
                    plunger_prep_pos=instrument.plunger_positions.bottom,
                    plunger_currents={
                        Axis.of_plunger(mount): instrument.plunger_motor_current.run
                    },
                    presses=[
                        PickUpTipPressSpec(
                            current={
                                Axis.by_mount(
                                    mount
                                ): instrument.pick_up_configurations.press_fit.current_by_tip_count[
                                    instrument.nozzle_manager.current_configuration.tip_count
                                ]
                            },
                            speed=pick_up_speed,
                            relative_down=top_types.Point(0, 0, press_dist),
                            relative_up=top_types.Point(0, 0, backup_dist),
                        )
                        for press_dist, backup_dist in build_presses()
                    ],
                    shake_off_list=self._build_pickup_shakes(instrument),
                    retract_target=instrument.pick_up_configurations.press_fit.distance
                    + check_incr * checked_presses
                    + 2,
                ),
                add_tip_to_instr,
            )
        else:
            return (
                PickUpTipSpec(
                    plunger_prep_pos=instrument.plunger_positions.bottom,
                    plunger_currents={
                        Axis.of_main_tool_actuator(
                            mount
                        ): instrument.plunger_motor_current.run
                    },
                    presses=[
                        PickUpTipPressSpec(
                            current={
                                Axis.by_mount(
                                    mount
                                ): instrument.pick_up_configurations.press_fit.current_by_tip_count[
                                    instrument.nozzle_manager.current_configuration.tip_count
                                ]
                            },
                            speed=pick_up_speed,
                            relative_down=top_types.Point(0, 0, press_dist),
                            relative_up=top_types.Point(0, 0, backup_dist),
                        )
                        for press_dist, backup_dist in build_presses()
                    ],
                    shake_off_list=self._build_pickup_shakes(instrument),
                    retract_target=instrument.pick_up_configurations.press_fit.distance
                    + check_incr * checked_presses
                    + 2,
                ),
                add_tip_to_instr,
            )

    @staticmethod
    def _shake_off_tips_drop(
        tiprack_diameter: float,
    ) -> List[Tuple[top_types.Point, Optional[float]]]:
        # tips don't always fall off, especially if resting against
        # tiprack or other tips below it. To ensure the tip has fallen
        # first, shake the pipette to dislodge partially-sealed tips,
        # then second, raise the pipette so loosened tips have room to fall
        shake_off_dist = SHAKE_OFF_TIPS_DROP_DISTANCE
        if tiprack_diameter > 0.0:
            shake_off_dist = min(shake_off_dist, tiprack_diameter / 4)
        shake_off_dist = max(shake_off_dist, 1.0)
        speed = SHAKE_OFF_TIPS_SPEED
        return [
            (top_types.Point(-shake_off_dist, 0, 0), speed),  # left
            (top_types.Point(2 * shake_off_dist, 0, 0), speed),  # right
            (top_types.Point(-shake_off_dist, 0, 0), speed),  # center
            (top_types.Point(0, 0, DROP_TIP_RELEASE_DISTANCE), None),  # top
        ]

    def _droptip_sequence_builder(
        self,
        bottom_pos: float,
        droptip_pos: float,
        plunger_currents: Dict[Axis, float],
        drop_tip_currents: Dict[Axis, float],
        speed: float,
        home_after: bool,
        home_axes: Sequence[Axis],
    ) -> Callable[[], List[DropTipMove]]:
        def build() -> List[DropTipMove]:
            base = [
                DropTipMove(
                    target_position=bottom_pos, current=plunger_currents, speed=None
                ),
                DropTipMove(
                    target_position=droptip_pos,
                    current=drop_tip_currents,
                    speed=speed,
                    home_after=home_after,
                    home_after_safety_margin=abs(bottom_pos - droptip_pos),
                    home_axes=home_axes,
                ),
                DropTipMove(  # always finish drop-tip at a known safe plunger position
                    target_position=bottom_pos, current=plunger_currents, speed=None
                ),
            ]
            return base

        return build

    @overload
    def plan_check_drop_tip(
        self, mount: top_types.Mount, home_after: bool
    ) -> Tuple[DropTipSpec, Callable[[], None]]:
        ...

    @overload
    def plan_check_drop_tip(
        self, mount: OT3Mount, home_after: bool
    ) -> Tuple[DropTipSpec, Callable[[], None]]:
        ...

    def plan_check_drop_tip(  # type: ignore[no-untyped-def]
        self,
        mount,
        home_after,
    ):
        instrument = self.get_pipette(mount)

        if not instrument.drop_configurations.plunger_eject:
            raise CommandPreconditionViolated(
                f"Pipette {instrument.name} on {mount.name} has no plunger eject configuration"
            )
        bottom = instrument.plunger_positions.bottom
        droptip = instrument.plunger_positions.drop_tip
        speed = instrument.drop_configurations.plunger_eject.speed
        shakes: List[Tuple[top_types.Point, Optional[float]]] = []
        if Quirks.dropTipShake in instrument.config.quirks:
            diameter = instrument.current_tiprack_diameter
            shakes = self._shake_off_tips_drop(diameter)

        def _remove_tips() -> None:
            instrument.set_current_volume(0)
            instrument.current_tiprack_diameter = 0.0
            instrument.remove_tip()

        if isinstance(mount, top_types.Mount):
            seq_builder_ot2 = self._droptip_sequence_builder(
                bottom,
                droptip,
                {Axis.of_plunger(mount): instrument.plunger_motor_current.run},
                {
                    Axis.of_plunger(
                        mount
                    ): instrument.drop_configurations.plunger_eject.current
                },
                speed,
                home_after,
                (Axis.of_plunger(mount),),
            )
            seq_ot2 = seq_builder_ot2()
            if Quirks.doubleDropTip in instrument.config.quirks:
                seq_ot2 = seq_ot2 + seq_builder_ot2()
            return (
                DropTipSpec(
                    drop_moves=seq_ot2,
                    shake_moves=shakes,
                    ending_current={
                        Axis.of_plunger(mount): instrument.plunger_motor_current.run
                    },
                ),
                _remove_tips,
            )
        else:
            seq_builder_ot3 = self._droptip_sequence_builder(
                bottom,
                droptip,
                {
                    Axis.of_main_tool_actuator(
                        mount
                    ): instrument.plunger_motor_current.run
                },
                {
                    Axis.of_main_tool_actuator(
                        mount
                    ): instrument.drop_configurations.plunger_eject.current
                },
                speed,
                home_after,
                (Axis.of_main_tool_actuator(mount),),
            )

            seq_ot3 = seq_builder_ot3()
            if Quirks.doubleDropTip in instrument.config.quirks:
                seq_ot3 = seq_ot3 + seq_builder_ot3()
            return (
                DropTipSpec(
                    drop_moves=seq_ot3,
                    shake_moves=shakes,
                    ending_current={
                        Axis.of_main_tool_actuator(
                            mount
                        ): instrument.plunger_motor_current.run
                    },
                ),
                _remove_tips,
            )

    def get_pipette(self, mount: MountType) -> Pipette:
        pip = self._attached_instruments[mount]
        if not pip:
            raise top_types.PipetteNotAttachedError(
                f"No pipette attached to {mount.name} mount"
            )
        return pip

    async def set_liquid_class(self, mount: MountType, liquid_class: str) -> None:
        pip = self.get_pipette(mount)
        pip.set_liquid_class_by_name(liquid_class)


class OT3PipetteHandler(PipetteHandlerProvider[OT3Mount]):
    """Override for correct plunger_position."""

    def plunger_position(
        self, instr: Pipette, ul: float, action: "UlPerMmAction"
    ) -> float:
        mm = ul / instr.ul_per_mm(ul, action)
        position = instr.plunger_positions.bottom - mm
        return round(position, 6)

    def critical_point_for(
        self, mount: MountType, cp_override: Optional[CriticalPoint] = None
    ) -> top_types.Point:
        pip = self._attached_instruments[OT3Mount.from_mount(mount)]
        if pip is not None and cp_override != CriticalPoint.MOUNT:
            return pip.critical_point(cp_override)
        else:
            return top_types.Point(0, 0, 0)



"""ProtocolEngine-based Well core implementations."""
from typing import Optional

from opentrons_shared_data.labware.constants import WELL_NAME_PATTERN

from opentrons.protocol_engine import WellLocation, WellOrigin, WellOffset
from opentrons.protocol_engine.clients import SyncClient as EngineClient
from opentrons.protocols.api_support.util import APIVersionError
from opentrons.types import Point

from . import point_calculations
from . import stringify
from ..well import AbstractWellCore
from ..._liquid import Liquid


class WellCore(AbstractWellCore):
    """Well API core using a ProtocolEngine.

    Args:
        name: The well's name in the labware, e.g. `A1`.
        labware_id: The ProtocolEngine ID of the well's parent labware.
        engine_client: Synchronous ProtocolEngine client.
    """

    def __init__(self, name: str, labware_id: str, engine_client: EngineClient) -> None:
        self._labware_id = labware_id
        self._engine_client = engine_client
        self._definition = engine_client.state.labware.get_well_definition(
            labware_id=labware_id, well_name=name
        )

        name_match = WELL_NAME_PATTERN.match(name)
        self._name = name
        self._row_name = name_match.group(1) if name_match is not None else ""
        self._column_name = name_match.group(2) if name_match is not None else ""

    @property
    def labware_id(self) -> str:
        """Get the ID of the well's parent labware."""
        return self._labware_id

    @property
    def diameter(self) -> Optional[float]:
        """Get the well's diameter, if circular."""
        return self._definition.diameter

    @property
    def length(self) -> Optional[float]:
        """Get the well's length, if rectangular."""
        return self._definition.xDimension

    @property
    def width(self) -> Optional[float]:
        """Get the well's width, if rectangular."""
        return self._definition.yDimension

    @property
    def depth(self) -> float:
        """Get the well's depth."""
        return self._definition.depth

    def has_tip(self) -> bool:
        """Whether the well contains a tip."""
        return self._engine_client.state.tips.has_clean_tip(
            self._labware_id, self._name
        )

    def set_has_tip(self, value: bool) -> None:
        """Set the well as containing or not containing a tip."""
        raise APIVersionError(
            "Manually setting the tip state of a well in a tip rack has been deprecated."
        )

    def get_display_name(self) -> str:
        """Get the full display name of the well (e.g. "A1 of Some Labware on 5")."""
        return stringify.well(
            engine_client=self._engine_client,
            well_name=self._name,
            labware_id=self._labware_id,
        )

    def get_name(self) -> str:
        """Get the name of the well (e.g. "A1")."""
        return self._name

    def get_column_name(self) -> str:
        """Get the column portion of the well name (e.g. "1")."""
        return self._column_name

    def get_row_name(self) -> str:
        """Get the row portion of the well name (e.g. "A")."""
        return self._row_name

    def get_max_volume(self) -> float:
        """Get the well's maximum liquid volume."""
        return self._definition.totalLiquidVolume

    def get_top(self, z_offset: float) -> Point:
        """Get the coordinate of the well's top, with a z-offset."""
        return self._engine_client.state.geometry.get_well_position(
            well_name=self._name,
            labware_id=self._labware_id,
            well_location=WellLocation(
                origin=WellOrigin.TOP, offset=WellOffset(x=0, y=0, z=z_offset)
            ),
        )

    def get_bottom(self, z_offset: float) -> Point:
        """Get the coordinate of the well's bottom, with a z-offset."""
        return self._engine_client.state.geometry.get_well_position(
            well_name=self._name,
            labware_id=self._labware_id,
            well_location=WellLocation(
                origin=WellOrigin.BOTTOM, offset=WellOffset(x=0, y=0, z=z_offset)
            ),
        )

    def get_center(self) -> Point:
        """Get the coordinate of the well's center."""
        return self._engine_client.state.geometry.get_well_position(
            well_name=self._name,
            labware_id=self._labware_id,
            well_location=WellLocation(origin=WellOrigin.CENTER),
        )

    def load_liquid(
        self,
        liquid: Liquid,
        volume: float,
    ) -> None:
        """Load liquid into a well."""
        self._engine_client.load_liquid(
            labware_id=self._labware_id,
            liquid_id=liquid._id,
            volume_by_well={self._name: volume},
        )

    def from_center_cartesian(self, x: float, y: float, z: float) -> Point:
        """Gets point in deck coordinates based on percentage of the radius of each axis."""
        well_size = self._engine_client.state.labware.get_well_size(
            labware_id=self.labware_id, well_name=self._name
        )

        return point_calculations.get_relative_offset(
            point=self.get_center(),
            size=well_size,
            x_ratio=x,
            y_ratio=y,
            z_ratio=z,
        )



""" opentrons.util.entrypoint_util: functions common to entrypoints
"""

import asyncio
import contextlib
from dataclasses import dataclass
import json
import logging
from json import JSONDecodeError
import pathlib
import tempfile
from typing import (
    Dict,
    Generator,
    List,
    Optional,
    Sequence,
    Union,
    TYPE_CHECKING,
)

from jsonschema import ValidationError  # type: ignore

from opentrons.config import IS_ROBOT, JUPYTER_NOTEBOOK_LABWARE_DIR
from opentrons.protocol_api import labware
from opentrons.calibration_storage import helpers
from opentrons.protocol_engine.errors.error_occurrence import (
    ErrorOccurrence as ProtocolEngineErrorOccurrence,
)
from opentrons.protocol_engine.types import DeckConfigurationType
from opentrons.protocol_reader import ProtocolReader, ProtocolSource
from opentrons.protocols.types import JsonProtocol, Protocol, PythonProtocol

if TYPE_CHECKING:
    from opentrons_shared_data.labware.dev_types import LabwareDefinition


log = logging.getLogger(__name__)


@dataclass
class FoundLabware:
    """An individual labware found by `labware_from_paths()`."""

    path: pathlib.Path
    definition: "LabwareDefinition"


def labware_from_paths(
    paths: Sequence[Union[str, pathlib.Path]]
) -> Dict[str, FoundLabware]:
    """Search paths for labware definitions.

    Returns:
        A dict, keyed by labware URI, where each value has the file path and the parsed def.
    """
    labware_defs: Dict[str, FoundLabware] = {}

    for strpath in paths:
        log.info(f"local labware: checking path {strpath}")
        purepath = pathlib.PurePath(strpath)
        if purepath.is_absolute():
            path = pathlib.Path(purepath)
        else:
            path = pathlib.Path.cwd() / purepath
        if not path.is_dir():
            raise RuntimeError(f"{path} is not a directory")
        for child in path.iterdir():
            if child.is_file() and child.suffix.endswith("json"):
                try:
                    defn = labware.verify_definition(child.read_bytes())
                except (ValidationError, JSONDecodeError):
                    log.info(f"{child}: invalid labware, ignoring")
                    log.debug(
                        f"{child}: labware invalid because of this exception.",
                        exc_info=True,
                    )
                else:
                    uri = helpers.uri_from_definition(defn)
                    labware_defs[uri] = FoundLabware(path=child, definition=defn)
                    log.info(f"loaded labware {uri} from {child}")
            else:
                log.info(f"ignoring {child} in labware path")
    return labware_defs


def find_jupyter_labware() -> Optional[Dict[str, FoundLabware]]:
    """Return labware files in this robot's Jupyter Notebook directory.

    Returns:
        If we're running on an Opentrons robot:
        A dict, keyed by labware URI, where each value has the file path and the parsed def.

        Otherwise: None.
    """
    if IS_ROBOT:
        # JUPYTER_NOTEBOOK_LABWARE_DIR should never be None when IS_ROBOT == True.
        assert JUPYTER_NOTEBOOK_LABWARE_DIR is not None
        if JUPYTER_NOTEBOOK_LABWARE_DIR.is_dir():
            return labware_from_paths([JUPYTER_NOTEBOOK_LABWARE_DIR])

    return None


def datafiles_from_paths(paths: Sequence[Union[str, pathlib.Path]]) -> Dict[str, bytes]:
    datafiles: Dict[str, bytes] = {}
    for strpath in paths:
        log.info(f"data files: checking path {strpath}")
        purepath = pathlib.PurePath(strpath)
        if purepath.is_absolute():
            path = pathlib.Path(purepath)
        else:
            path = pathlib.Path.cwd() / purepath
        if path.is_file():
            datafiles[path.name] = path.read_bytes()
            log.info(f"read {path} into custom data as {path.name}")
        elif path.is_dir():
            for child in path.iterdir():
                if child.is_file():
                    datafiles[child.name] = child.read_bytes()
                    log.info(f"read {child} into data path as {child.name}")
                else:
                    log.info(f"ignoring {child} in data path")
    return datafiles


def get_deck_configuration() -> DeckConfigurationType:
    """Return the host robot's current deck configuration."""
    # TODO: Search for the file where robot-server stores it.
    # Flex: /var/lib/opentrons-robot-server/deck_configuration.json
    # OT-2: /data/opentrons_robot_server/deck_configuration.json
    # https://opentrons.atlassian.net/browse/RSS-400
    return []


@contextlib.contextmanager
def adapt_protocol_source(protocol: Protocol) -> Generator[ProtocolSource, None, None]:
    """Convert a `Protocol` to a `ProtocolSource`.

    `Protocol` and `ProtocolSource` do basically the same thing. `Protocol` is the traditional
    interface. `ProtocolSource` is the newer, but not necessarily better, interface that's required
    to run stuff through Protocol Engine. Ideally, the two would be unified. Until then, we have
    this shim.

    This is a context manager because it needs to keep some temp files around.
    """
    # ProtocolReader needs to know the filename of the main protocol file so it can infer from its
    # extension whether it's a JSON or Python protocol. But that filename doesn't necessarily exist,
    # like when a user passes a text stream to opentrons.simulate.simulate(). As a hack, work
    # backwards and synthesize a dummy filename with the correct extension.
    if protocol.filename is not None:
        # We were given a filename, so no need to guess.
        #
        # It's not well-defined in our customer-facing interfaces whether the supplied protocol_name
        # should be just the filename part, or a path with separators. In case it contains stuff
        # like "../", sanitize it to just the filename part so we don't save files somewhere bad.
        main_file_name = pathlib.Path(protocol.filename).name
    elif isinstance(protocol, JsonProtocol):
        main_file_name = "protocol.json"
    else:
        main_file_name = "protocol.py"

    with tempfile.TemporaryDirectory() as temporary_directory:
        # FIXME(mm, 2023-06-26): Copying these files is pure overhead, and it introduces encoding
        # hazards. Remove this when we can parse JSONv6+ and PAPIv2.14+ protocols without going
        # through the filesystem. https://opentrons.atlassian.net/browse/RSS-281

        main_file = pathlib.Path(temporary_directory) / main_file_name
        if isinstance(protocol.text, str):
            main_file.write_text(protocol.text, encoding="utf-8")
        else:
            main_file.write_bytes(protocol.text)

        labware_files: List[pathlib.Path] = []
        if isinstance(protocol, PythonProtocol) and protocol.extra_labware is not None:
            for labware_index, labware_definition in enumerate(
                protocol.extra_labware.values()
            ):
                new_labware_file = (
                    pathlib.Path(temporary_directory) / f"{labware_index}.json"
                )
                new_labware_file.write_text(
                    json.dumps(labware_definition), encoding="utf-8"
                )
                labware_files.append(new_labware_file)

        protocol_source = asyncio.run(
            ProtocolReader().read_saved(
                files=[main_file] + labware_files,
                directory=None,
                files_are_prevalidated=False,
            )
        )

        yield protocol_source


class ProtocolEngineExecuteError(Exception):
    def __init__(self, errors: List[ProtocolEngineErrorOccurrence]) -> None:
        """Raised when there was any fatal error running a protocol through Protocol Engine.

        Protocol Engine reports errors as data, not as exceptions.
        But the only way for `opentrons.execute.execute()` and `opentrons.simulate.simulate()`
        to signal problems to their callers is to raise something.
        So we need this class to wrap them.

        Params:
            errors: The errors that Protocol Engine reported.
        """
        # Show the full error details if this is part of a traceback. Don't try to summarize.
        super().__init__(errors)

        self._error_occurrences = errors

    def to_stderr_string(self) -> str:
        """Return a string suitable as the stderr output of the `opentrons_execute` CLI.

        This summarizes from the full error details.
        """
        # It's unclear what exactly we should extract here.
        #
        # First, do we print the first element, or the last, or all of them?
        #
        # Second, do we print the .detail? .errorCode? .errorInfo? .wrappedErrors?
        # By contract, .detail seems like it would be insufficient, but experimentally,
        # it includes a lot, like:
        #
        #     ProtocolEngineError [line 3]: Error 4000 GENERAL_ERROR (ProtocolEngineError):
        #     UnexpectedProtocolError: Labware "fixture_12_trough" not found with version 1
        #     in namespace "fixture".
        return self._error_occurrences[0].detail



"""Magnetic Module protocol commands."""

from .disengage import (
    Disengage,
    DisengageCreate,
    DisengageParams,
    DisengageResult,
    DisengageCommandType,
)
from .engage import (
    Engage,
    EngageCreate,
    EngageParams,
    EngageResult,
    EngageCommandType,
)


__all__ = [
    # magneticModule/disengageMagnet
    "Disengage",
    "DisengageCreate",
    "DisengageParams",
    "DisengageResult",
    "DisengageCommandType",
    # magneticModule/engageMagnet
    "Engage",
    "EngageCreate",
    "EngageParams",
    "EngageResult",
    "EngageCommandType",
]



"""Touch tip command request, result, and implementation models."""
from __future__ import annotations
from pydantic import Field
from typing import TYPE_CHECKING, Optional, Type
from typing_extensions import Literal

from ..errors import TouchTipDisabledError, LabwareIsTipRackError
from ..types import DeckPoint
from .command import AbstractCommandImpl, BaseCommand, BaseCommandCreate
from .pipetting_common import (
    PipetteIdMixin,
    WellLocationMixin,
    DestinationPositionResult,
)

if TYPE_CHECKING:
    from ..execution import MovementHandler, GantryMover
    from ..state import StateView


TouchTipCommandType = Literal["touchTip"]


class TouchTipParams(PipetteIdMixin, WellLocationMixin):
    """Payload needed to touch a pipette tip the sides of a specific well."""

    radius: float = Field(
        1.0,
        description=(
            "The proportion of the target well's radius the pipette tip will move towards."
        ),
    )

    speed: Optional[float] = Field(
        None,
        description=(
            "Override the travel speed in mm/s."
            " This controls the straight linear speed of motion."
        ),
    )


class TouchTipResult(DestinationPositionResult):
    """Result data from the execution of a TouchTip."""

    pass


class TouchTipImplementation(AbstractCommandImpl[TouchTipParams, TouchTipResult]):
    """Touch tip command implementation."""

    def __init__(
        self,
        state_view: StateView,
        movement: MovementHandler,
        gantry_mover: GantryMover,
        **kwargs: object,
    ) -> None:
        self._state_view = state_view
        self._movement = movement
        self._gantry_mover = gantry_mover

    async def execute(self, params: TouchTipParams) -> TouchTipResult:
        """Touch tip to sides of a well using the requested pipette."""
        pipette_id = params.pipetteId
        labware_id = params.labwareId
        well_name = params.wellName

        if self._state_view.labware.get_has_quirk(labware_id, "touchTipDisabled"):
            raise TouchTipDisabledError(
                f"Touch tip not allowed on labware {labware_id}"
            )

        if self._state_view.labware.is_tiprack(labware_id):
            raise LabwareIsTipRackError("Cannot touch tip on tip rack")

        center_point = await self._movement.move_to_well(
            pipette_id=pipette_id,
            labware_id=labware_id,
            well_name=well_name,
            well_location=params.wellLocation,
        )

        touch_speed = self._state_view.pipettes.get_movement_speed(
            pipette_id, params.speed
        )

        touch_waypoints = self._state_view.motion.get_touch_tip_waypoints(
            pipette_id=pipette_id,
            labware_id=labware_id,
            well_name=well_name,
            radius=params.radius,
            center_point=center_point,
        )

        x, y, z = await self._gantry_mover.move_to(
            pipette_id=pipette_id,
            waypoints=touch_waypoints,
            speed=touch_speed,
        )

        return TouchTipResult(position=DeckPoint(x=x, y=y, z=z))


class TouchTip(BaseCommand[TouchTipParams, TouchTipResult]):
    """Touch up tip command model."""

    commandType: TouchTipCommandType = "touchTip"
    params: TouchTipParams
    result: Optional[TouchTipResult]

    _ImplementationCls: Type[TouchTipImplementation] = TouchTipImplementation


class TouchTipCreate(BaseCommandCreate[TouchTipParams]):
    """Touch tip command creation request model."""

    commandType: TouchTipCommandType = "touchTip"
    params: TouchTipParams

    _CommandCls: Type[TouchTip] = TouchTip



"""Action pipeline module."""
from typing import List

from .action_handler import ActionHandler
from .actions import Action


class ActionDispatcher:
    """A pipeline, with an endpoint, that actions can be dispatched into."""

    def __init__(self, sink: ActionHandler) -> None:
        """Initialize the ActionDispatcher and action pipeline.

        Arguments:
            sink: The action handler that all actions in the pipeline
                are sent to.
        """
        self._handlers: List[ActionHandler] = []
        self._sink = sink

    def add_handler(self, handler: ActionHandler) -> None:
        """Add an action handler to the pipeline before the sink."""
        self._handlers.append(handler)

    def dispatch(self, action: Action) -> None:
        """Dispatch an action into the pipeline."""
        for handler in self._handlers:
            handler.handle_action(action)

        self._sink.handle_action(action)



"""Pipette config data providers."""
from dataclasses import dataclass
from typing import Dict, Optional

from opentrons_shared_data.pipette.dev_types import PipetteName, PipetteModel
from opentrons_shared_data.pipette import (
    pipette_load_name_conversions as pipette_load_name,
    load_data as load_pipette_data,
    types as pip_types,
    pipette_definition,
)


from opentrons.hardware_control.dev_types import PipetteDict
from opentrons.hardware_control.nozzle_manager import (
    NozzleConfigurationManager,
    NozzleMap,
)

from ..types import FlowRates
from ...types import Point


@dataclass(frozen=True)
class LoadedStaticPipetteData:
    """Static pipette config data for load pipette."""

    model: str
    display_name: str
    min_volume: float
    max_volume: float
    channels: int
    home_position: float
    nozzle_offset_z: float
    flow_rates: FlowRates
    tip_configuration_lookup_table: Dict[
        float, pipette_definition.SupportedTipsDefinition
    ]
    nominal_tip_overlap: Dict[str, float]
    back_left_corner_offset: Point
    front_right_corner_offset: Point
    back_left_nozzle_offset: Point
    front_right_nozzle_offset: Point


class VirtualPipetteDataProvider:
    """Provide pipette data without requiring hardware control."""

    def __init__(self) -> None:
        """Build a VirtualPipetteDataProvider."""
        self._liquid_class_by_id: Dict[str, pip_types.LiquidClasses] = {}
        self._nozzle_manager_layout_by_id: Dict[str, NozzleConfigurationManager] = {}

    def configure_virtual_pipette_nozzle_layout(
        self,
        pipette_id: str,
        pipette_model_string: str,
        back_left_nozzle: Optional[str] = None,
        front_right_nozzle: Optional[str] = None,
        starting_nozzle: Optional[str] = None,
    ) -> None:
        """Emulate update_nozzle_configuration_for_mount."""
        if pipette_id not in self._nozzle_manager_layout_by_id:
            config = self._get_virtual_pipette_full_config_by_model_string(
                pipette_model_string
            )
            new_nozzle_manager = NozzleConfigurationManager.build_from_config(config)
            if back_left_nozzle and front_right_nozzle:
                new_nozzle_manager.update_nozzle_configuration(
                    back_left_nozzle, front_right_nozzle, starting_nozzle
                )
            self._nozzle_manager_layout_by_id[pipette_id] = new_nozzle_manager
        elif back_left_nozzle and front_right_nozzle:
            # Need to make sure that we pass all the right nozzles here.
            self._nozzle_manager_layout_by_id[pipette_id].update_nozzle_configuration(
                back_left_nozzle, front_right_nozzle, starting_nozzle
            )
        else:
            self._nozzle_manager_layout_by_id[
                pipette_id
            ].reset_to_default_configuration()

    def configure_virtual_pipette_for_volume(
        self, pipette_id: str, volume: float, pipette_model_string: str
    ) -> None:
        """Emulate configure_for_volume with the same logic for changing modes."""
        if pipette_id not in self._liquid_class_by_id:
            self._liquid_class_by_id[pipette_id] = pip_types.LiquidClasses.default
        pipette_model = pipette_load_name.convert_pipette_model(
            PipetteModel(pipette_model_string)
        )
        config = load_pipette_data.load_definition(
            pipette_model.pipette_type,
            pipette_model.pipette_channels,
            pipette_model.pipette_version,
        )

        liquid_class = pipette_definition.liquid_class_for_volume_between_default_and_defaultlowvolume(
            volume, self._liquid_class_by_id[pipette_id], config.liquid_properties
        )
        self._liquid_class_by_id[pipette_id] = liquid_class

    def get_nozzle_layout_for_pipette(self, pipette_id: str) -> NozzleMap:
        """Get the current nozzle layout stored for a virtual pipette."""
        return self._nozzle_manager_layout_by_id[pipette_id].current_configuration

    def get_virtual_pipette_static_config_by_model_string(
        self, pipette_model_string: str, pipette_id: str
    ) -> LoadedStaticPipetteData:
        """Get the config of a pipette when you know its model string (e.g. from state)."""
        pipette_model = pipette_load_name.convert_pipette_model(
            PipetteModel(pipette_model_string)
        )
        return self._get_virtual_pipette_static_config_by_model(
            pipette_model, pipette_id
        )

    def _get_virtual_pipette_full_config_by_model_string(
        self, pipette_model_string: str
    ) -> pipette_definition.PipetteConfigurations:
        """Get the full pipette config from a model string."""
        pipette_model = pipette_load_name.convert_pipette_model(
            PipetteModel(pipette_model_string)
        )
        return load_pipette_data.load_definition(
            pipette_model.pipette_type,
            pipette_model.pipette_channels,
            pipette_model.pipette_version,
        )

    def _get_virtual_pipette_static_config_by_model(
        self, pipette_model: pipette_definition.PipetteModelVersionType, pipette_id: str
    ) -> LoadedStaticPipetteData:
        if pipette_id not in self._liquid_class_by_id:
            self._liquid_class_by_id[pipette_id] = pip_types.LiquidClasses.default

        liquid_class = self._liquid_class_by_id[pipette_id]
        config = load_pipette_data.load_definition(
            pipette_model.pipette_type,
            pipette_model.pipette_channels,
            pipette_model.pipette_version,
        )
        try:
            tip_type = pip_types.PipetteTipType(
                config.liquid_properties[liquid_class].max_volume
            )
        except ValueError:
            tip_type = pipette_definition.default_tip_for_liquid_class(
                config.liquid_properties[liquid_class]
            )
        tip_configuration = config.liquid_properties[liquid_class].supported_tips[
            tip_type
        ]

        nozzle_manager = NozzleConfigurationManager.build_from_config(config)
        pip_back_left = config.pipette_bounding_box_offsets.back_left_corner
        pip_front_right = config.pipette_bounding_box_offsets.front_right_corner
        return LoadedStaticPipetteData(
            model=str(pipette_model),
            display_name=config.display_name,
            min_volume=config.liquid_properties[liquid_class].min_volume,
            max_volume=config.liquid_properties[liquid_class].max_volume,
            channels=config.channels,
            home_position=config.mount_configurations.homePosition,
            nozzle_offset_z=config.nozzle_offset[2],
            tip_configuration_lookup_table={
                k.value: v
                for k, v in config.liquid_properties[
                    liquid_class
                ].supported_tips.items()
            },
            flow_rates=FlowRates(
                default_blow_out=tip_configuration.default_blowout_flowrate.values_by_api_level,
                default_aspirate=tip_configuration.default_aspirate_flowrate.values_by_api_level,
                default_dispense=tip_configuration.default_dispense_flowrate.values_by_api_level,
            ),
            nominal_tip_overlap=config.liquid_properties[
                liquid_class
            ].tip_overlap_dictionary,
            back_left_corner_offset=Point(
                pip_back_left[0], pip_back_left[1], pip_back_left[2]
            ),
            front_right_corner_offset=Point(
                pip_front_right[0], pip_front_right[1], pip_front_right[2]
            ),
            back_left_nozzle_offset=nozzle_manager.current_configuration.back_left_nozzle_offset,
            front_right_nozzle_offset=nozzle_manager.current_configuration.front_right_nozzle_offset,
        )

    def get_virtual_pipette_static_config(
        self, pipette_name: PipetteName, pipette_id: str
    ) -> LoadedStaticPipetteData:
        """Get the config for a virtual pipette, given only the pipette name."""
        pipette_model = pipette_load_name.convert_pipette_name(pipette_name)
        return self._get_virtual_pipette_static_config_by_model(
            pipette_model, pipette_id
        )


def get_pipette_static_config(pipette_dict: PipetteDict) -> LoadedStaticPipetteData:
    """Get the config for a pipette, given the state/config object from the HW API."""
    back_left_offset = pipette_dict["pipette_bounding_box_offsets"].back_left_corner
    front_right_offset = pipette_dict["pipette_bounding_box_offsets"].front_right_corner
    return LoadedStaticPipetteData(
        model=pipette_dict["model"],
        display_name=pipette_dict["display_name"],
        min_volume=pipette_dict["min_volume"],
        max_volume=pipette_dict["max_volume"],
        channels=pipette_dict["channels"],
        flow_rates=FlowRates(
            default_blow_out=pipette_dict["default_blow_out_flow_rates"],
            default_aspirate=pipette_dict["default_aspirate_flow_rates"],
            default_dispense=pipette_dict["default_dispense_flow_rates"],
        ),
        tip_configuration_lookup_table={
            k.value: v for k, v in pipette_dict["supported_tips"].items()
        },
        nominal_tip_overlap=pipette_dict["tip_overlap"],
        # TODO(mc, 2023-02-28): these two values are not present in PipetteDict
        # https://opentrons.atlassian.net/browse/RCORE-655
        home_position=0,
        nozzle_offset_z=0,
        # TODO (spp): Confirm that the nozzle map is populated by the hardware api by default
        back_left_nozzle_offset=pipette_dict[
            "current_nozzle_map"
        ].back_left_nozzle_offset,
        front_right_nozzle_offset=pipette_dict[
            "current_nozzle_map"
        ].front_right_nozzle_offset,
        back_left_corner_offset=Point(
            back_left_offset[0], back_left_offset[1], back_left_offset[2]
        ),
        front_right_corner_offset=Point(
            front_right_offset[0], front_right_offset[1], front_right_offset[2]
        ),
    )



from abc import ABC, abstractmethod

from opentrons.types import Location, Point
from opentrons_shared_data.labware.dev_types import LabwareDefinition


class AbstractLabwareGeometry(ABC):
    """Convenience interface for accessing geometry properties of a labware."""

    @property
    @abstractmethod
    def parent(self) -> Location:
        ...

    @property
    @abstractmethod
    def offset(self) -> Point:
        ...

    @property
    @abstractmethod
    def x_dimension(self) -> float:
        ...

    @property
    @abstractmethod
    def y_dimension(self) -> float:
        ...

    @property
    @abstractmethod
    def z_dimension(self) -> float:
        ...


class LabwareGeometry(AbstractLabwareGeometry):
    def __init__(self, definition: LabwareDefinition, parent: Location):
        """Constructor"""
        self._parent = parent
        offset = definition["cornerOffsetFromSlot"]
        self._offset = Point(offset["x"], offset["y"], offset["z"]) + parent.point

        dimensions = definition["dimensions"]
        self._x_dimension = dimensions["xDimension"]
        self._y_dimension = dimensions["yDimension"]
        self._z_dimension = dimensions["zDimension"]

    @property
    def parent(self) -> Location:
        return self._parent

    @property
    def offset(self) -> Point:
        return self._offset

    @property
    def x_dimension(self) -> float:
        return self._x_dimension

    @property
    def y_dimension(self) -> float:
        return self._y_dimension

    @property
    def z_dimension(self) -> float:
        return self._z_dimension



from typing import List
from typing_extensions import Protocol

from ..modules import AbstractModule, ModuleModel


class ModuleProvider(Protocol):
    """A protocol specifying access to modules."""

    @property
    def attached_modules(self) -> List[AbstractModule]:
        """Return a list of currently-attached modules."""
        ...

    async def create_simulating_module(self, model: ModuleModel) -> AbstractModule:
        """Create a simulating module hardware interface."""



from typing import Dict, Optional

from .abstract import AbstractMagDeckDriver
from opentrons.util.async_helpers import ensure_yield


MAG_DECK_MODELS = {
    "magneticModuleV1": "mag_deck_v1.1",
    "magneticModuleV2": "mag_deck_v20",
}


class SimulatingDriver(AbstractMagDeckDriver):
    def __init__(self, sim_model: Optional[str] = None) -> None:
        self._height = 0.0
        self._model = MAG_DECK_MODELS[sim_model] if sim_model else "mag_deck_v1.1"

    @ensure_yield
    async def probe_plate(self) -> None:
        pass

    @ensure_yield
    async def home(self) -> None:
        pass

    @ensure_yield
    async def move(self, location: float) -> None:
        self._height = location

    @ensure_yield
    async def get_device_info(self) -> Dict[str, str]:
        return {
            "serial": "dummySerialMD",
            "model": self._model,
            "version": "dummyVersionMD",
        }

    @ensure_yield
    async def connect(self) -> None:
        pass

    @ensure_yield
    async def disconnect(self) -> None:
        pass

    @ensure_yield
    async def enter_programming_mode(self) -> None:
        pass

    @ensure_yield
    async def is_connected(self) -> bool:
        return True

    @ensure_yield
    async def get_plate_height(self) -> float:
        return self._height

    @ensure_yield
    async def get_mag_position(self) -> float:
        return self._height



"""Hash command params into idempotent keys to track commands from analysis to run."""
from hashlib import md5
from typing import Optional

from .command import CommandIntent
from .command_unions import CommandCreate


# TODO(mm, 2023-04-28):
# This implementation will not notice that commands are different if they have different params
# but share the same commandType. We should also hash command params. (Jira RCORE-326.)
def hash_command_params(
    create: CommandCreate, last_hash: Optional[str]
) -> Optional[str]:
    """Given a command create object, return a hash.

    The hash is based on three things:
    - The command parameters
    - The latest hash (yo dawg I heard you like blockchains)
    - Whether the command is a setup command or a protocol command;
      setup commands are not hashed

    Args:
        create: The command create request.
        last_hash: The last command hash, if it exists.

    Returns:
        The command hash, if the command is a protocol command.
        `None` if the command is a setup command.
    """
    if create.intent == CommandIntent.SETUP:
        return None
    else:
        # We avoid Python's built-in hash() function because it's not stable across
        # runs of the Python interpreter. (Jira RSS-215.)
        last_contribution = b"" if last_hash is None else last_hash.encode("ascii")
        this_contribution = md5(create.commandType.encode("ascii")).digest()
        to_hash = last_contribution + this_contribution
        return md5(to_hash).hexdigest()



"""ProtocolEngine action interfaces.

Actions are the driver of state changes in the ProtocolEngine.
"""

from .action_dispatcher import ActionDispatcher
from .action_handler import ActionHandler
from .actions import (
    Action,
    PlayAction,
    PauseAction,
    PauseSource,
    StopAction,
    FinishAction,
    HardwareStoppedAction,
    QueueCommandAction,
    UpdateCommandAction,
    FailCommandAction,
    AddLabwareOffsetAction,
    AddLabwareDefinitionAction,
    AddLiquidAction,
    AddAddressableAreaAction,
    AddModuleAction,
    FinishErrorDetails,
    DoorChangeAction,
    ResetTipsAction,
    SetPipetteMovementSpeedAction,
)

__all__ = [
    # action pipeline interface
    "ActionDispatcher",
    # action reaction interface
    "ActionHandler",
    # action values
    "Action",
    "PlayAction",
    "PauseAction",
    "StopAction",
    "FinishAction",
    "HardwareStoppedAction",
    "QueueCommandAction",
    "UpdateCommandAction",
    "FailCommandAction",
    "AddLabwareOffsetAction",
    "AddLabwareDefinitionAction",
    "AddLiquidAction",
    "AddAddressableAreaAction",
    "AddModuleAction",
    "DoorChangeAction",
    "ResetTipsAction",
    "SetPipetteMovementSpeedAction",
    # action payload values
    "PauseSource",
    "FinishErrorDetails",
]



"""Retract Axis command payload, result, and implementation models."""
from __future__ import annotations
from pydantic import BaseModel, Field
from typing import TYPE_CHECKING, Optional, Type
from typing_extensions import Literal

from ..types import MotorAxis
from .command import AbstractCommandImpl, BaseCommand, BaseCommandCreate

if TYPE_CHECKING:
    from ..execution import MovementHandler

RetractAxisCommandType = Literal["retractAxis"]


class RetractAxisParams(BaseModel):
    """Payload required for a Retract Axis command."""

    axis: MotorAxis = Field(
        ...,
        description=(
            "Axis to retract to its home position as quickly as safely possible."
            " The difference between retracting an axis and homing an axis using the"
            " home command is that a home will always probe the limit switch and"
            " will work as the first motion command a robot will need to execute;"
            " On the other hand, retraction will rely on this previously determined "
            " home position to move to it as fast as safely possible."
            " So on the Flex, it will move (fast) the axis to the previously recorded home position"
            " and on the OT2, it will move (fast) the axis a safe distance from the previously"
            " recorded home position, and then slowly approach the limit switch."
        ),
    )


class RetractAxisResult(BaseModel):
    """Result data from the execution of a Rectract Axis command."""


class RetractAxisImplementation(
    AbstractCommandImpl[RetractAxisParams, RetractAxisResult]
):
    """Retract Axis command implementation."""

    def __init__(self, movement: MovementHandler, **kwargs: object) -> None:
        self._movement = movement

    async def execute(self, params: RetractAxisParams) -> RetractAxisResult:
        """Retract the specified axis."""
        await self._movement.retract_axis(axis=params.axis)
        return RetractAxisResult()


class RetractAxis(BaseCommand[RetractAxisParams, RetractAxisResult]):
    """Command to retract the specified axis to its home position."""

    commandType: RetractAxisCommandType = "retractAxis"
    params: RetractAxisParams
    result: Optional[RetractAxisResult]

    _ImplementationCls: Type[RetractAxisImplementation] = RetractAxisImplementation


class RetractAxisCreate(BaseCommandCreate[RetractAxisParams]):
    """Data to create a Retract Axis command."""

    commandType: RetractAxisCommandType = "retractAxis"
    params: RetractAxisParams

    _CommandCls: Type[RetractAxis] = RetractAxis



"""Magnetic Module engage command request, result, and implementation models."""
from __future__ import annotations
from typing import Optional, TYPE_CHECKING
from typing_extensions import Literal, Type

from pydantic import BaseModel, Field

from ..command import AbstractCommandImpl, BaseCommand, BaseCommandCreate

if TYPE_CHECKING:
    from opentrons.protocol_engine.execution import EquipmentHandler
    from opentrons.protocol_engine.state import StateView


EngageCommandType = Literal["magneticModule/engage"]


class EngageParams(BaseModel):
    """Input data to engage a Magnetic Module."""

    moduleId: str = Field(
        ...,
        description=(
            "The ID of the Magnetic Module whose magnets you want to raise,"
            " from a prior `loadModule` command."
        ),
    )

    height: float = Field(
        ...,
        description=(
            "How high, in millimeters, to raise the magnets."
            "\n\n"
            "Zero means the tops of the magnets are level with the ledge"
            " that the labware rests on."
            " This will be slightly above the magnets' minimum height,"
            " the hardware home position."
            " Negative values are allowed, to put the magnets below the ledge."
            "\n\n"
            "Units are always true millimeters."
            " This is unlike certain labware definitions,"
            " engage commands in the Python Protocol API,"
            " and engage commands in older versions of the JSON protocol schema."
            " Take care to convert properly."
        ),
    )


class EngageResult(BaseModel):
    """The result of a Magnetic Module engage command."""

    pass


class EngageImplementation(AbstractCommandImpl[EngageParams, EngageResult]):
    """The implementation of a Magnetic Module engage command."""

    def __init__(
        self,
        state_view: StateView,
        equipment: EquipmentHandler,
        **unused_dependencies: object,
    ) -> None:
        self._state_view = state_view
        self._equipment = equipment

    async def execute(self, params: EngageParams) -> EngageResult:
        """Execute a Magnetic Module engage command.

        Raises:
            ModuleNotLoadedError: If the given module ID doesn't point to a
                module that's already been loaded.
            WrongModuleTypeError: If the given module ID points to a non-Magnetic
                module.
            ModuleNotAttachedError: If the given module ID points to a valid loaded
                Magnetic Module, but that module's hardware wasn't found attached.
            EngageHeightOutOfRangeError: If the given height is unreachable.
        """
        # Allow propagation of ModuleNotLoadedError and WrongModuleTypeError.
        mag_module_substate = self._state_view.modules.get_magnetic_module_substate(
            module_id=params.moduleId
        )
        # Allow propagation of EngageHeightOutOfRangeError.
        hardware_height = mag_module_substate.calculate_magnet_hardware_height(
            mm_from_base=params.height,
        )
        # Allow propagation of ModuleNotAttachedError.
        hardware_module = self._equipment.get_module_hardware_api(
            mag_module_substate.module_id
        )

        if hardware_module is not None:  # Not virtualizing modules.
            await hardware_module.engage(height=hardware_height)

        return EngageResult()


class Engage(BaseCommand[EngageParams, EngageResult]):
    """A command to engage a Magnetic Module's magnets."""

    commandType: EngageCommandType = "magneticModule/engage"
    params: EngageParams
    result: Optional[EngageResult]

    _ImplementationCls: Type[EngageImplementation] = EngageImplementation


class EngageCreate(BaseCommandCreate[EngageParams]):
    """A request to create a Magnetic Module engage command."""

    commandType: EngageCommandType = "magneticModule/engage"
    params: EngageParams

    _CommandCls: Type[Engage] = Engage



from typing import Any, Dict, NamedTuple, Optional, Union, TYPE_CHECKING
from dataclasses import dataclass

from opentrons_shared_data.robot.dev_types import RobotType
from .api_support.definitions import MIN_SUPPORTED_VERSION
from .api_support.types import APIVersion

if TYPE_CHECKING:
    from opentrons_shared_data.labware.dev_types import LabwareDefinition
    from opentrons_shared_data.protocol.dev_types import (
        JsonProtocol as JsonProtocolDef,
        Metadata as JsonProtocolMetadata,
    )


PythonProtocolMetadata = Optional[Dict[str, str]]
"""The contents of a Python protocol's `metadata` dict, if there is one."""


PythonProtocolRequirements = Optional[Dict[str, str]]
"""The contents of a Python protocol's `requirements` dict, if there is one."""


@dataclass(frozen=True)
class StaticPythonInfo:
    """Information statically extracted from a Python Protocol API file."""

    metadata: PythonProtocolMetadata
    requirements: PythonProtocolRequirements


@dataclass(frozen=True)
class _ProtocolCommon:
    text: Union[str, bytes]
    """The original text of the protocol file in the format it was specified with.

    This leads to a wide type but it is actually quite important that we do not ever
    str.decode('utf-8') this because it will break the interpreter's understanding of
    line numbers for if we have to format an exception.
    """

    filename: Optional[str]
    """The original name of the main protocol file, if it had a name.

    For JSON protocols, this will be the name of the .json file.
    For Python protocols, this will be the name of the .py file.
    For bundled protocols, this will be the name of the .zip file.

    This can be `None` if, for example, we've parsed the protocol from an in-memory text stream.
    """

    # TODO(mm, 2023-06-22): Move api_level out of _ProtocolCommon and into PythonProtocol.
    # JSON protocols do not have an intrinsic api_level, especially since JSONv6,
    # where they are no longer executed via the Python Protocol API.
    api_level: "APIVersion"

    robot_type: RobotType


@dataclass(frozen=True)
class JsonProtocol(_ProtocolCommon):
    schema_version: int
    contents: "JsonProtocolDef"
    metadata: "JsonProtocolMetadata"


@dataclass(frozen=True)
class PythonProtocol(_ProtocolCommon):
    contents: Any  # This is the output of compile() which we can't type
    metadata: PythonProtocolMetadata
    # these 'bundled_' attrs should only be included when the protocol is a zip
    bundled_labware: Optional[Dict[str, "LabwareDefinition"]]
    bundled_data: Optional[Dict[str, bytes]]
    bundled_python: Optional[Dict[str, str]]
    # this should only be included when the protocol is not a zip
    extra_labware: Optional[Dict[str, "LabwareDefinition"]]


Protocol = Union[JsonProtocol, PythonProtocol]


class BundleContents(NamedTuple):
    protocol: Union[str, bytes]
    bundled_labware: Dict[str, "LabwareDefinition"]
    bundled_data: Dict[str, bytes]
    bundled_python: Dict[str, str]


RUN_FUNCTION_MESSAGE = """\
A Python protocol must define a function called 'run' that takes a
single argument: the protocol context to call functions on. For instance, a run
function might look like this:

def run(ctx):
    ctx.comment('hello, world')

This function is called by the robot when the robot executes the protocol.
"""

PYTHON_API_VERSION_DEPRECATED = """

The Python protocol you uploaded has the Python API Version {0}.  Robot server version 4.0.0 is
the official end of life of Python API Version {0}. The minimum supported Python API Version is {1}. This means that this protocol
will not run in robot server version 4.0.0 and above.
Please downgrade your robot server version if you wish to run this protocol. Otherwise, please upgrade this
protocol to Python API Version {1} or above.

To upgrade your protocol to Python API Version {1} or above, please view our documentation at https://docs.opentrons.com/v2/index.html.

Please contact support@opentrons.com to retrieve the previous software version and be guided
through the downgrade process.

"""


# TODO(mm, 2023-08-07): Align with the app team on how to surface errors like this,
# and probably make this an EnumeratedError.
class MalformedPythonProtocolError(Exception):
    def __init__(
        self, short_message: str, long_additional_message: Optional[str] = None
    ) -> None:
        """Raised when a user's Python protocol file is malformed.

        "Malformed" in this case means it's either syntactically invalid as standard Python, or it
        doesn't conform to the Python Protocol API's structural requirements, like having a `run()`
        function and declaring an `apiLevel`. This does not cover runtime or analysis errors such as
        aspirating from a nonexistent well or running out of tips.

        Params:
            short_message: A short (~1-2 sentences), self-contained message describing what's wrong
                with the file. This should not contain internal newlines or indentation.

            long_additional_message: Longer context, in addition to `short_message`. This may
                contain internal newlines and indentation.
        """
        self.short_message = short_message
        self.long_additional_message = long_additional_message
        super().__init__(short_message, long_additional_message)

    def __str__(self) -> str:
        if self.long_additional_message is not None:
            return self.short_message + "\n\n" + self.long_additional_message
        else:
            return self.short_message

    def __repr__(self) -> str:
        return "<{}: {}>".format(self.__class__.__name__, self.short_message)


class ApiDeprecationError(Exception):
    def __init__(self, version: APIVersion) -> None:
        self.version = version
        super().__init__(version)

    def __str__(self) -> str:
        return PYTHON_API_VERSION_DEPRECATED.format(self.version, MIN_SUPPORTED_VERSION)

    def __repr__(self) -> str:
        return "<{}: {}>".format(self.__class__.__name__, self.version)
